import math
import warnings
from typing import Callable, List, Optional, Sequence, Tuple
from dataclasses import dataclass

import eikonalfm
import matplotlib.pyplot as plt
import numpy as np
from skimage import measure

from mesh import Mesh
from util import arrayutils
# local installation
try:
    from lib_amplitude import amplitude
# system-wide installation
except ModuleNotFoundError:
    import amplitude


@dataclass
class GRTData:
    dx: Tuple[float, float]
    phi: np.ndarray
    integrand_factor: np.ndarray

def calculate_ta_data(c: np.ndarray, dx: Tuple[float, float], fs_tol: Optional[float]=1e-3) -> Tuple[np.ndarray, np.ndarray]:
    """
    Calculates traveltime and amplitude data.

    Parameters
    ----------
        c : ndarray
            the background velocity (needs to be only depth dependent)
        dx : 2-Tuple of floats
            the grid spacing of c
        fs_tol : float (optional)
            tolerance for the fast sweeping algorithm (amplitude)

    Returns
    ----------
        tau : np.ndarray
            traveltime
        a : np.ndarray
            amplitude divided by c
    """
    # solve eikonal equation from origin
    tau0 = eikonalfm.distance(c.shape, dx, (0, 0), indexing="ij")
    tau1 = eikonalfm.factored_fast_marching(c, (0, 0), dx, 2)

    # solve transport equation from origin
    a0 = amplitude.base_amplitude(c.shape, dx, (0, 0), indexing="ij")
    a1, _ = amplitude.factored_amplitude(c, tau1, (0, 0), dx, fs_tol)
    # already divide a by c here for A = 1/c^2...
    return tau0 * tau1, a0*a1 / c

def calculate_GRT_data(c: np.ndarray, dx: Tuple[float, float], tau: np.ndarray, a: np.ndarray, alpha: float) -> GRTData:
    """
    Calculates the required data to realize the linear operator F for the equation F*n = y from given traveltime and amplitude data.

    Parameters
    ----------
        c : ndarray
            the background velocity (needs to be only depth dependent)
        dx : 2-Tuple of floats
            the grid spacing of c
        tau : np.ndarray
            traveltime (see `calculate_ta_data`)
        a : np.ndarray
            amplitude divided by c
        alpha : float
            offset for the common-offset geometry

    Returns
    ----------
        grt_data : GRTData
            data object containing phi, integral factor and dx. Needed to evaluate F*n.
    """
    assert(c.shape == tau.shape == a.shape), "shapes do not match"
    # common-offset geometry
    # we always ceil alpha and floor s (see below)
    alpha_ind = math.ceil(alpha / dx[0])

    # tau for source and receiver (needed for gradients later)
    tau_source = tau[alpha_ind:, :]
    tau_recv = np.vstack((tau[alpha_ind:0:-1, :], tau[:tau.shape[0]-2*alpha_ind, :]))
    phi = tau_source + tau_recv

    # dot product of gradients of tau (needed later for the integrand)
    tau_grad_dot = arrayutils.dot(np.gradient(tau_source, *dx, edge_order=2), np.gradient(tau_recv, *dx, edge_order=2))

    # A as product of a_source and a_receiver
    A = a[alpha_ind:, :] * np.vstack((a[alpha_ind:0:-1, :], a[:a.shape[0]-2*alpha_ind, :]))

    # slice c to match the other array shapes
    c = c[alpha_ind:, :]

    # integrand = c * A * n / np.sqrt(1 + c**2 * tau_grad_dot)
    # do this calculation here so we only need 2 shared nparray instead of 4
    integrand_factor = c * A / np.sqrt(1 + c**2 * tau_grad_dot)
    return GRTData(dx, phi, integrand_factor)

def calculate_F_data(c: np.ndarray, dx: Tuple[float, float], alpha: float, fs_tol: Optional[float]=1e-3) -> GRTData:
    """
    Calculates all required data to realize the linear operator F for the equation F*n = g.

    Parameters
    ----------
        c : ndarray
            the background velocity (needs to be only depth dependent)
        dx : 2-Tuple of floats
            the grid spacing of c
        alpha : float
            offset for the common-offset geometry
        fs_tol : float (optional)
            tolerance for the fast sweeping algorithm (amplitude)

    Returns
    ----------
        grt_data : GRTData
            data object containing phi, integral factor and dx. Needed to evaluate F*n.
    """
    tau, a = calculate_ta_data(c, dx, fs_tol=fs_tol)
    return calculate_GRT_data(c, dx, tau, a, alpha)


def __quadrature(points: np.ndarray, values: np.ndarray) -> float:
    """
    Trapezoidal rule
    """
    assert len(points) == len(values)

    # distances between each point
    dist = np.linalg.norm(points[1:] - points[:-1], axis=1, ord=2)
    # sum of two adjacent values
    value_sum = values[:-1] + values[1:]
    return np.sum(value_sum * dist) / 2


_ISQRT2 = 1 / np.sqrt(2)
def F_tilde(grt_data: GRTData, y_min: float, s_list: Sequence[float], t: float, n_func: Callable[[np.ndarray, Sequence[float]], np.ndarray], *n_args, res: Optional[np.ndarray]=None) -> Sequence[float]:
    """
    Applies the operator F for the common offset geometry, i.e. source at (s - alpha, y_min), receiver at (s + alpha, y_min).

    Parameters
    ----------
        grt_data : GRTData
            data object containing phi, integral factor and dx
        y_min: float
            upper bound of the domain
        s_list : sequence of floats
            list of all offsets
        t : float
            time at which the contours are calculated
        n_func : function
            mollifier (kernel) or n (phantom) function
        n_args : any
            additional arguments to "n_func"
        res: ndarray (optional)
            arrray to write results into

    Returns
    ----------
        res : ndarray
            array containing the values of F*n for each offset in "s"
    """
    if res is not None:
        assert len(res) == len(s_list), f"length of result array ({len(res)}) doesn't match number of s-values ({len(s_list)})"
        # assume that res is zeroed
    else:
        res = [0] * len(s_list)

    # Find contours at a constant value of t
    # TODO: maybe implement own marching squares which always starts the search at the top (faster?)
    contours = measure.find_contours(grt_data.phi, t)

    if not contours:
        t_min, t_max = grt_data.phi.min(), grt_data.phi.max()
        raise Exception("No contours found for t={:.2f}. Range is [{:.2f}, {:.2f}]".format(t, t_min, t_max))
    elif len(contours) > 2:
        warnings.warn("more than two contours found for t={:.2f}".format(t), Warning)
        plt.figure()
        plt.suptitle(f"t={t}")
        plt.gca().invert_yaxis()
        plt.gca().set_aspect("equal")
        plt.pcolormesh(grt_data.phi.T, cmap="gray")
        for cont in contours:
            plt.plot(cont[:, 0], cont[:, 1])
        plt.show(block=True)

    # integrate over reflection isochrones by using the contour points as indices
    # L_(s, t) = {x in X: phi(s, x) = t}
    # L_(s, t) = L_(0, t) + (s, 0)^T
    for points in contours:
        # contour does not always start/end in x=0 or end/start in y=0
        if points[-1, 0] != 0:
            if points[0, 0] != 0:
                warnings.warn("Contour for t={:.2f} does not start or end in x=0. Maybe it lies outside of the domain.".format(t), Warning)
            elif points[-1, 1] != 0:
                warnings.warn("Contour for t={:.2f} does not end in y=0. Maybe it lies outside of the domain.".format(t), Warning)
        elif points[0, 1] != 0:
            warnings.warn("Contour for t={:.2f} does not start in y=0. Maybe it lies outside of the domain.".format(t), Warning)

        # real points without offset 's'
        # x-coord of the first point is greater than the last one -> contour is clock-wise
        if points[0, 0] > points[-1, 0]:
            real_points_base = np.vstack((points, (-1, 1) * points[-2::-1])) * grt_data.dx
            # use original points for array access
            points = np.vstack((points, points[-2::-1]))
        # contour is counter clock-wise
        else:
            real_points_base = np.vstack((points[::-1], (-1, 1) * points[1:])) * grt_data.dx
            # use original points for array access
            points = np.vstack((points[::-1], points[1:]))

        # TODO: maybe split the method here (and return points and real_points_base)
        # this would enable to calculate the points for a single t and use them for all p and s

        for i, s in enumerate(s_list):
            real_points = real_points_base + (s, y_min)

            n = n_func(real_points, *n_args)
            n_ind = n.nonzero()
            # ignore points where the mollifier is zero for calculating the integrand
            points_nonzero = points[n_ind]

            integrand = np.zeros(len(points))
            #integrand[n_ind] = exact(2.0, s, points_nonzero) * n[n_ind]
            integrand[n_ind] = arrayutils.interp_index(points_nonzero, grt_data.integrand_factor) * n[n_ind]

            # apply quadrature rule
            res[i] += _ISQRT2 * __quadrature(points, integrand)

    return res


def check_t_range(c: np.ndarray, mesh: Mesh, alpha: float, t: Sequence[float], expand_domain=True) -> None:
    """
    Checks if the mesh will be big enough to contain all reflection isochrones for t.

    Parameters
    ----------
        c : ndarray
            (only) depth-dependant background velocity
        mesh : Mesh
            the domain mesh of c, starting at x=0.
        alpha: float
            offset for the common-offset geometry
        t : sequence of floats
            list of all time steps
        expand_domain : bool (optional)
            whether c should be expanded by alpha to the right
    """
    assert c.shape == mesh.n, "shapes of c and the mesh do not match."
    assert mesh.domain[0, 0] == 0, "x-start of mesh must be 0"
    # we need to add the lenth of alpha to the first axis of the mesh
    n_alpha = math.ceil(alpha / mesh.dx[0])

    if expand_domain:
        n_total = mesh.n[0] + n_alpha

        c_stack = []
        # It's possible that n_total > c.shape[0], so we handle that case first
        while n_total > c.shape[0]:
            c_stack.append(c)
            n_total -= c.shape[0]
        # now add the remaining width and stack c along the x1-axis
        c_stack.append(c[:n_total])
        c_ref = np.vstack(c_stack)
    else:
        c_ref = c

    # solve eikonal equation from origin
    tau0 = eikonalfm.distance(c_ref.shape, mesh.dx, (0, 0), indexing="ij")
    tau1 = eikonalfm.factored_fast_marching(c_ref, (0, 0), mesh.dx, 2)
    tau = tau0 * tau1

    # sum of tau for source and receiver
    phi = tau[n_alpha:, :] + np.vstack((tau[n_alpha:0:-1, :], tau[:tau.shape[0]-2*n_alpha, :]))

    def min_cont_cond(contours):
        if not (1 <= len(contours) <= 2):
            return False

        # x-coordinate of first or last point must be 0
        # y-coordinate of last or first point must be 0
        if not ((contours[0][0, 0] == 0 and contours[0][-1, 1] == 0) or (contours[0][-1, 0] == 0 and contours[0][0, 1] == 0)):
            return False
        if len(contours) == 2 and not ((contours[1][0, 0] == 0 and contours[1][-1, 1] == 0) or (contours[1][-1, 0] == 0 and contours[1][0, 1] == 0)):
            return False

        return True

    t_tmp, t_min = phi.min(), phi[n_alpha + 1, 0]
    # assume that contours for 't_min' fulfill the 'min_cont_cond'
    contours_min = measure.find_contours(phi, t_min)
    assert min_cont_cond(contours_min)

    def plot_contours(contours, label, color):
        for cont in contours:
            plt.plot(cont[:, 0], cont[:, 1], color=color)
        # handle for the legend
        plt.plot([], [], label=label, color=color)

    # find the minimal possible contours
    while True:
        if t_min - t_tmp < mesh.dx[0] / 20:
            break

        # binary search
        t_mid = (t_min + t_tmp) / 2
        contours = measure.find_contours(phi, t_mid)

        if min_cont_cond(contours):
            t_min = t_mid
            contours_min = contours
        else:
            t_tmp = t_mid

    t_max = min(np.min(phi[:, -1]), np.min(phi[-1, :]))
    contours_max: List[np.ndarray] = []
    while len(contours_max) == 0:
        t_max -= 1e-3
        contours_max = measure.find_contours(phi, t_max)

    if t[0] < t_min:
        print("[Warning] t[0]={} could be too small, t_min={}.".format(t[0], t_min))
    if t[-1] > t_max:
        print("[Warning] t[-1]={} could be too large, t_max={}. Decrease t or make the domain bigger.".format(t[-1], t_max))

    def plot_contours(contours, label, color):
        for cont in contours:
            plt.plot(cont[:, 0], cont[:, 1], color=color)
        # handle for the legend
        plt.plot([], [], label=label, color=color)

    plt.figure()
    plt.suptitle(r"Your $t = [{}, {}]$".format(t[0], t[-1]))
    plt.gca().invert_yaxis()
    plt.gca().set_aspect("equal")
    plt.contourf(phi.T, cmap="gray", levels=100)
    plot_contours(contours_min, r"$t_{min} = " +"{:.4f}".format(t_min)+ r"$", color="blue")
    contours_t0 = measure.find_contours(phi, t[0])
    plot_contours(contours_t0, r"$t_0 = " +"{:.4f}".format(t[0])+ r"$", color="green")
    contours_t1 = measure.find_contours(phi, t[-1])
    plot_contours(contours_t1, r"$t_{-1} = " +"{:.4f}".format(t[-1])+ r"$", color="orange")
    plot_contours(contours_max, r"$t_{max} = " +"{:.4f}".format(t_max)+ r"$", color="red")

    # change label text
    plt.gca().xaxis.set_major_formatter(plt.FuncFormatter(lambda x, pos: f"{x * mesh.dx[0]:.1f}"))
    plt.gca().yaxis.set_major_formatter(plt.FuncFormatter(lambda x, pos: f"{mesh.domain.start[1] + x * mesh.dx[1]:.1f}"))
    plt.legend()
    plt.show()
