import os
import sys

import numpy as np
import scipy.sparse as sp
from matplotlib.colors import TwoSlopeNorm
import matplotlib.pyplot as plt

import GRT
import kernel
import rhs
import pseudo_ops
import reconstruction
from util import arrayutils, confighelper, plotutil, profiling, reshelper, testsetup


def calc_kernels(config: confighelper.Config):
    testsetup.init()

    mesh = config.geteval("DEFAULT", "mesh")
    print(mesh)

    # velocity
    def c_func(y):
        res = config.getlambda("DEFAULT", "c", "y")(y)
        if np.isscalar(res):
            return res * np.ones(y.shape)
        else:
            return res
    c = c_func(mesh.X[1])

    alpha = config.getfloat("DEFAULT", "alpha")
    t = config.geteval("DEFAULT", "t")
    s = config.geteval("DEFAULT", "s")

    # mesh for the reconstruction
    p_mesh = config.geteval("DEFAULT", "p_mesh")

    p_ref = np.linspace(*p_mesh.domain[1], num=config.getint("DEFAULT", "n_ref_kernels"))

    fs_tol = config.getfloat("DEFAULT", "fs_tol")

    # pseudo-diff operator
    sec = config["PSEUDO_OP"]
    pseudo_op_type = sec.get("type")
    if pseudo_op_type == "x2_Laplace":
        pseudo_op = pseudo_ops.x2_Laplace(sec.getfloat("k"), sec.getfloat("gamma"), power=sec.getfloat("power"))
    elif pseudo_op_type == "Laplace":
        pseudo_op = pseudo_ops.Laplace(sec.getfloat("k"), sec.getfloat("gamma"))
    else:
        raise Exception(pseudo_op_type + " is not a valid pseudo_op type")

    # profiling.start()
    timer = profiling.Timer()
    kernels_ref, _, grt_data = kernel.calculate_reference_kernels(mesh, p_mesh.domain, c, alpha, s, t, p_ref, pseudo_op, fs_tol=fs_tol)
    print(timer)

    path = reshelper.next_file(reshelper.res_path("saved", config.get("DEFAULT", "folder")), f"kernels.pcl")
    name = os.path.basename(os.path.splitext(path)[0])
    reshelper.save_data(os.path.join(config.get("DEFAULT", "folder"), name), {
        'kernels': [sp.csr_matrix(K) for K in kernels_ref],
        'grt_data': grt_data,
        'config': config
    })

    return int(name[len("kernels"):])


def _check_config(data, config):
    config_gen = data['config']

    # everything except "n_func", "noise_eps", "cutoff" must match
    for section in config_gen:
        for k, v in config_gen[section].items():
            # if config[section][k] != v:
            if config[section][k] != v and k not in ["n_func", "noise_eps", "cutoff"]:
                raise Exception(f"""[Warning] current config and config for kernel-calculation do not match. Kernels were generated with
    {k} = {v}, but now
    {k} = {config[section][k]}. Aborting...""")

def reconstruct(config: confighelper.Config, num: int):    
    # load and check data
    path = os.path.join(config.get("DEFAULT", "folder"), f"kernels{num}")
    data = reshelper.load_data(path)
    _check_config
    
    testsetup.init()
    mesh = config.geteval("DEFAULT", "mesh")

    t = config.geteval("DEFAULT", "t")
    s = config.geteval("DEFAULT", "s")

    # calculate the cutoff radon transform
    n_func = config.getnfunc("DEFAULT", "n_func")
    rt = rhs.calculate_rt(mesh, data['grt_data'], s, t, n_func)

    # apply cutoff function
    cutoff_conf = config.geteval("DEFAULT", "cutoff")
    cutoff_map = lambda val, cut: None if cut is None else (val + cut[0], val + cut[1])        
    cutoff_params = tuple((cutoff_map(val, cut) for val, cut in zip((s[0], s[-1], t[0], t[-1]), cutoff_conf)))
    cutoff_rt = rhs.cut_off(s, t, *cutoff_params) * rt

    # add noise
    noise_eps = config.getfloat("DEFAULT", "noise_eps", fallback=0.0)
    if noise_eps != 0:
        from numpy.random import default_rng
        rng = default_rng()
        cutoff_rt = arrayutils.add_noise(cutoff_rt, rng, noise_eps)

    p_mesh = config.geteval("DEFAULT", "p_mesh")
    p_ref = np.linspace(*p_mesh.domain[1], num=config.getint("DEFAULT", "n_ref_kernels"))

    # L2-skalar product of cuttoff-data and reconstruction kernel
    res = reconstruction.scalar_prod(p_mesh, s, t, p_ref, data['kernels'], cutoff_rt)


    # save data and create overview pdf
    foldername = config.get("DEFAULT", "folder")
    name = "rec" + ("noised" if noise_eps > 0 else "") + f"{num}_.pdf"
    fn = reshelper.next_file(reshelper.fig_path(foldername), name)
    fn = os.path.splitext(fn)[0]
    reshelper.save_data(os.path.join(foldername, os.path.basename(fn)), {
        'rec': res,
        'g': cutoff_rt,
        'config': config
    })


    mesh = config.geteval("DEFAULT", "mesh")

    # velocity
    def c_func(y):
        res = config.getlambda("DEFAULT", "c", "y")(y)
        if np.isscalar(res):
            return res * np.ones(y.shape)
        else:
            return res
    c = c_func(mesh.X[1])

    alpha = config.getfloat("DEFAULT", "alpha")
    fs_tol = config.getfloat("DEFAULT", "fs_tol")
    
    # pseudo-diff operator
    sec = config["PSEUDO_OP"]
    if sec.get("type") == "x2_Laplace":
        pseudo_op_str = r"\Delta x_2^{" + str(sec.getfloat("power") or 1) + r"}"
    else:
        pseudo_op_str = "\Delta"

    # reconstruction image
    fig, ax = testsetup.create_figure()
    ax.set_xlabel("$p_1$")
    ax.set_ylabel("$p_2$")
    plot = ax.pcolormesh(*p_mesh.X, res, cmap="gray", shading="auto", rasterized=True)
    plotutil.add_colorbar(ax, plot)
    fig.savefig(fn + ".pdf")
    plt.close(fig)

    # rhs
    fig, ax = testsetup.create_figure()
    ax.set_xlabel("$s$")
    ax.set_ylabel("$t$")
    plot = ax.pcolormesh(s, t, cutoff_rt.T, cmap="Greys", rasterized=True, shading="auto")
    plotutil.add_colorbar(ax, plot)
    fig.savefig(fn + "_g.pdf")
    plt.close(fig)

    # phantom n
    fig, ax = testsetup.create_figure()
    ax.set_xlabel("$p_1$")
    ax.set_ylabel("$p_2$")
    rhs.plot_n_func(p_mesh, n_func)
    fig.savefig(fn + "_n.pdf")
    plt.close(fig)

    # velocity image
    fig, ax = testsetup.create_figure()
    ax.set_xlabel("$x_1$")
    ax.set_ylabel("$x_2$")
    plot = ax.pcolormesh(*mesh.X, c, cmap="Greys", rasterized=True, shading="auto")
    plotutil.add_colorbar(ax, plot)
    fig.savefig(fn + "_c.pdf")
    plt.close(fig)
    
    print(f"Image saved to {fn}.pdf")


def main(config, num=None):
    print(f"running reconstruction for '{args.configfile}'")

    if num is None:
        num = calc_kernels(config)

    reconstruct(config, num)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("configfile", help="path to the .ini file for the reconstruction")
    parser.add_argument("num", type=int, nargs="?", default=None, help='the corresponding data generation number (e.g. for "g1.pcl" it is 1')
    args = parser.parse_args()
    # check if config file is valid
    if not args.configfile.endswith(".ini") or not os.path.exists(args.configfile):
        parser.print_help()
        sys.exit()

    config = confighelper.Config(args.configfile)

    main(config, args.num)
