# run this with MPI for better performance/parallelization
# mpiexec -n 2 python tests/inconsistent_data/pysit_data.py
import sys
import os
from typing import Sequence
import matplotlib.pyplot as plt
import numpy as np
import pysit
from mpi4py import MPI
import argparse

import rhs
from util import arrayutils, confighelper, plotutil, reshelper, profiling, testsetup


# plot configuration (debug)
plot_velocity = False
animate_wavefields = False
plot_seismograms = False
plot_difference = False
plot_integral_func = False

# setup config

parser = argparse.ArgumentParser()
parser.add_argument("configfile", help="path to the .ini file for the reconstruction")
args = parser.parse_args()
# check if config file is valid
if not args.configfile.endswith(".ini") or not os.path.exists(args.configfile):
    parser.print_help()
    sys.exit()
config = confighelper.Config(args.configfile)

comm = MPI.COMM_WORLD

# offsets for data generation
s_data = config.geteval("GENERATION", "s")
# prevent the program to be run with too many processes
if comm.size > len(s_data) and comm.rank == 0:
    print("[Error] nprocs ({}) must be smaller than n_s ({})".format(comm.size, len(s_data)))
    comm.Abort()
comm.Barrier()

data_mesh = config.geteval("GENERATION", "mesh")
alpha = config.getfloat("DEFAULT", "alpha")

if (data_mesh.domain[0, 0] > s_data[0] - alpha or data_mesh.domain[0, 1] < s_data[-1] + alpha) and comm.rank == 0:
    print("[Error] [s_min-alpha, s_max+alpha] must be contained in data mesh")
    comm.Abort()
comm.Barrier()

if comm.rank == 0:
    print(f"running data generation for '{args.configfile}' with {comm.size} process(es)")
    timer = profiling.Timer()

# split s in to nearly equal parts for each process
s_splits = np.array_split(s_data, comm.size)
s_batch = s_splits[comm.rank]

# reference velocity
def c_func(y):
    res = config.getlambda("GENERATION", "c", "y")(y)
    if np.isscalar(res):
        return res * np.ones(y.shape)
    else:
        return res
c = c_func(data_mesh.X[1])

# real velocity
v = np.array(c, copy=True)
v_add = config.getvelocity("GENERATION", "v", data_mesh)
v /= np.sqrt(1 + v_add)

if plot_velocity and comm.rank == 0:
    # plot speed of sound v
    plt.figure()
    ax = testsetup.subplot(1, 2, 1)
    ax.set_title("true material $v$")
    plot = ax.pcolormesh(*data_mesh.X, v, cmap="Greys", shading="auto")
    plotutil.add_colorbar(ax, plot)

    ax = testsetup.subplot(1, 2, 2)
    ax.set_title("background material $c$")
    plot = ax.pcolormesh(*data_mesh.X, c, cmap="Greys", shading="auto")
    # plot.set_clim(v.min(), v.max())
    plotutil.add_colorbar(ax, plot)
    plt.show()

t_range = config.geteval("GENERATION", "t_range")

def solve(solver: rhs.ForwardSolver, shots: Sequence[pysit.Shot], velocity: np.ndarray, output_wavefields: bool=False):
    if output_wavefields:
        seismograms, t, wavefields = solver.solve(shots, t_range, velocity, output_wavefields=True)
    else:
        seismograms, t = solver.solve(shots, t_range, velocity, output_wavefields=False)
    seismograms = np.array(seismograms, copy=False, order="C") # ensure column-major ordering

    shapes = list(map(lambda x: len(x) * seismograms.shape[1], s_splits))
    displacements = np.cumsum([0] + shapes[:-1])

    if output_wavefields:
        return seismograms, t, wavefields, shapes, displacements
    else:
        return seismograms, t, shapes, displacements

pml = config.geteval("GENERATION", "pml")
fwd = rhs.ForwardSolver(data_mesh, pmlx=pml[0], pmly=pml[1])
shots = fwd.generate_shots(s_batch, alpha)

if animate_wavefields and comm.rank == 0:
    seismograms_batch, t, wavefields, shapes, displacements = solve(fwd, shots, v, output_wavefields=True)
    # plot wavefields (for first source)
    rhs.animate(wavefields[0], fwd.mesh, t, speed=0.5)
else:
    seismograms_batch, t, shapes, displacements = solve(fwd, shots, v)
seismograms_bg_batch, t_bg, shapes_bg, displacements_bg = solve(fwd, shots, c)

# collect all batches
seismograms = np.empty((len(s_data), seismograms_batch.shape[1]))
seismograms_bg = np.empty((len(s_data), seismograms_bg_batch.shape[1]))
comm.Barrier()
comm.Gatherv(seismograms_batch, [seismograms, shapes, displacements, MPI.DOUBLE])
comm.Gatherv(seismograms_bg_batch, [seismograms_bg, shapes_bg, displacements_bg, MPI.DOUBLE])


if plot_seismograms and comm.rank == 0:
    plt.figure()
    plt.suptitle("seismograms")

    t_min = config.getfloat("RECONSTRUCTION", "t_min")
    # scale seismograms to see them better
    scale = alpha / np.abs(seismograms[:, t >= 1.05 * t_min]).max()

    ax = plt.subplot(2, 1, 1)
    ax.set_title("real velocity")
    for i, seismo in enumerate(seismograms):
        ax.plot(t, s_data[i] + scale * seismo)

    ax = plt.subplot(2, 1, 2)
    ax.set_title("background velocity")
    for i, seismo in enumerate(seismograms_bg):
        ax.plot(t_bg, s_data[i] + scale * seismo)
    plt.show()

if comm.rank == 0:
    # normalize seismograms to the same t-shape
    tt, seismograms, seismograms_bg = rhs.interpolate_seismograms(t, seismograms, t_bg, seismograms_bg)
    seismograms = rhs.fit_seismograms(tt, seismograms, seismograms_bg)
    diff = seismograms_bg - seismograms
    # try to remove initial wave
    # diff *= rhs.cut_off_axis(tt, (0, 0.5), None)

    # plot seismogram difference
    if plot_difference:
        t_min = config.getfloat("RECONSTRUCTION", "t_min")
        t_flag = tt >= t_min
        scale = alpha / np.abs(diff).max()

        plt.figure()
        plt.suptitle("seismograms (difference)")

        ax = plt.subplot(2, 1, 1)
        for i, seismo in enumerate(diff):
            ax.plot(tt, s_data[i] + scale * seismo)

        scale = alpha / np.abs(diff[:, t_flag]).max()
        ax = plt.subplot(2, 1, 2)
        for i, seismo in enumerate(diff):
            ax.plot(tt[t_flag], s_data[i] + scale * seismo[t_flag])
        plt.show()

    # y(t) = 4 Pi * int_0^t (seismo_real - seismo_bg) dt
    data = 4*np.pi * arrayutils.integral_function(tt, diff, axis=1)

    if plot_integral_func:
        t_min = config.getfloat("RECONSTRUCTION", "t_min")
        t_flag = tt >= t_min
        scale = alpha / np.abs(data).max()

        # plot seismograms
        plt.figure()
        plt.suptitle("integral function")

        ax = plt.subplot(2, 1, 1)
        for i, seismo in enumerate(data):
            ax.plot(tt, s_data[i] + scale * seismo)

        scale = alpha / np.abs(data[:, t_flag]).max()
        ax = plt.subplot(2, 1, 2)
        for i, seismo in enumerate(data):
            ax.plot(tt[t_flag], s_data[i] + scale * seismo[t_flag])

        plt.show()

    print(timer)

    path = reshelper.next_file(reshelper.res_path("saved", config.get("DEFAULT", "folder")), f"g.pcl")
    name = os.path.basename(os.path.splitext(path)[0])
    reshelper.save_data(os.path.join(config.get("DEFAULT", "folder"), name), {
        'g': data,
        't': tt,
        'config': config
    })
