import sys
import os
import matplotlib.pyplot as plt
import numpy as np
import rhs
from util import confighelper, plotutil, reshelper, testsetup
import GRT
import kernel
import pseudo_ops
import reconstruction
import argparse


# plot configuration (debug)
plot_rhs = False


def _check_config(data, config):
    config_gen = data['config']

    for section in ["DEFAULT", "GENERATION"]:
        for k, v in config_gen[section].items():
            if config[section][k] != v:
                raise Exception(f"""[Warning] current config and config for data generation do not match. Data was generated with
    {k}={v}, but is now
    {k}={config[section][k]}. Aborting...""")


def do_reconstruction(num, config):
    # load and check data
    path = os.path.join(config.get("DEFAULT", "folder"), f"g{num}")
    data = reshelper.load_data(path)
    _check_config(data, config)

    testsetup.init()

    g = data['g']
    tt = data['t']
    s_data = config.geteval("GENERATION", "s")

    t_min = config.getfloat("RECONSTRUCTION", "t_min")
    g = g[:, tt >= t_min]
    t = tt[tt >= t_min]

    # apply cutoff function
    cutoff_conf = config.geteval("RECONSTRUCTION", "cutoff")
    cutoff_map = lambda val, cut: None if cut is None else (val + cut[0], val + cut[1])
    cutoff_params = tuple((cutoff_map(val, cut) for val, cut in zip((s_data[0], s_data[-1], t[0], t[-1]), cutoff_conf)))
    g *= rhs.cut_off(s_data, t, *cutoff_params)

    alpha = config.getfloat("DEFAULT", "alpha")
    if plot_rhs:
        plt.figure()
        plt.suptitle("(cutoff) integral function")

        ax = testsetup.subplot(1, 2, 1, invert_yaxis=False, equal=False)
        scale = alpha / np.abs(g).max()
        for i, seismo in enumerate(reversed(g)):
            ax.plot(t, s_data[len(g)-1 - i] + scale * seismo)
        
        ax = testsetup.subplot(1, 2, 2, equal=False)
        plot = ax.pcolormesh(s_data, t, g.T, cmap="Greys", rasterized=True, shading="auto")
        plotutil.add_colorbar(ax, plot)
        plt.show()


    mesh = config.geteval("RECONSTRUCTION", "mesh")

    # reconstruction velocity
    def c_func(y):
        res = config.getlambda("RECONSTRUCTION", "c", "y")(y)
        if np.isscalar(res):
            return res * np.ones(y.shape)
        else:
            return res
    c = c_func(mesh.X[1])


    p_mesh = config.geteval("RECONSTRUCTION", "p_mesh")
    print(p_mesh)

    # ymin, ymax = p_mesh.domain[1, 0], p_mesh.domain[1, 1]
    # p_num = max(min(2 * (ymax - ymin) / (config.getfloat("PSEUDO_OP", "gamma") / 2), p_mesh.n[1]), 2)
    p_ref = np.linspace(*p_mesh.domain[1], num=config.getint("RECONSTRUCTION", "n_ref_kernels"))
    print("#reference kernels: {}".format(len(p_ref)))

    s = config.geteval("RECONSTRUCTION", "s")


    # pseudo-diff operator
    sec = config["PSEUDO_OP"]
    pseudo_op_type = sec.get("type")
    if pseudo_op_type == "x2_Laplace":
        pseudo_op = pseudo_ops.x2_Laplace(sec.getfloat("k"), sec.getfloat("gamma"), power=sec.getfloat("power"))
    elif pseudo_op_type == "Laplace":
        pseudo_op = pseudo_ops.Laplace(sec.getfloat("k"), sec.getfloat("gamma"))
    else:
        raise Exception(pseudo_op_type + " is not a valid pseudo_op type")

    kernels_ref, _, _ = kernel.calculate_reference_kernels(mesh, p_mesh.domain, c, alpha, s, t, p_ref,
                                                            pseudo_op, fs_tol=config.getfloat("RECONSTRUCTION", "fs_tol"))

    # resize data to match kernels
    # import cv2
    # g = cv2.resize(g.T, (len(s), len(t)), interpolation=cv2.INTER_LANCZOS4).T

    # L2-skalar product of cuttoff-data and reconstruction kernel
    res = reconstruction.scalar_prod(p_mesh, s, t, p_ref, kernels_ref, g)


    data_mesh = config.geteval("GENERATION", "mesh")
    # reconstruction velocity
    def c_data_func(y):
        res = config.getlambda("GENERATION", "c", "y")(y)
        if np.isscalar(res):
            return res * np.ones(y.shape)
        else:
            return res
    c_data = c_data_func(data_mesh.X[1])

    # real velocity
    v = np.array(c_data, copy=True)
    v_add = config.getvelocity("GENERATION", "v", data_mesh)
    v /= np.sqrt(1 + v_add)

    fn = reshelper.next_file(reshelper.fig_path(config.get("DEFAULT", "folder")), f"rec{num}_.pdf")
    name = os.path.splitext(fn)[0]

    # reconstruction image
    fig, ax = testsetup.create_figure()
    ax.set_xlabel("$p_1$")
    ax.set_ylabel("$p_2$")
    plot = ax.pcolormesh(*p_mesh.X, res, cmap="gray", shading="auto", rasterized=True)
    plotutil.add_colorbar(ax, plot)
    fig.savefig(name + ".pdf")
    plt.close(fig)

    # rhs
    fig, ax = testsetup.create_figure()
    ax.set_xlabel("$s$")
    ax.set_ylabel("$t$")
    plot = ax.pcolormesh(s_data, t, g.T, cmap="Greys", rasterized=True, shading="auto")
    plotutil.add_colorbar(ax, plot)
    fig.savefig(name + "_g.pdf")
    plt.close(fig)

    # seismogram data plot (right-hand side)
    scale = alpha / np.abs(g).max()
    fig, ax = testsetup.create_figure(invert_yaxis=False, equal=False)
    for i, seismo in enumerate(reversed(g)):
        ax.plot(t, s_data[len(g)-1 - i] + scale * seismo)
    ax.set_xlabel("$t$")
    ax.set_ylabel("$s$")
    fig.savefig(name + "_seismograms.pdf")
    plt.close(fig)

    # real velocity plot
    fig, ax = testsetup.create_figure()
    ax.set_xlabel(r"$x_1$")
    ax.set_ylabel(r"$x_2$")
    plot = ax.pcolormesh(*data_mesh.X, v, cmap="gray", shading="auto", rasterized=True)
    plotutil.add_colorbar(ax, plot)
    fig.savefig(name + "_v.pdf")
    plt.close(fig)
    
    print(f"Image saved to {name}.pdf")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("configfile", help="path to the .ini file for the reconstruction")
    parser.add_argument("num", type=int, help='the corresponding data generation number (e.g. for "g1.pcl" it is 1')

    args = parser.parse_args()
    # check if config file is valid
    if not args.configfile.endswith(".ini") or not os.path.exists(args.configfile):
        parser.print_help()
        sys.exit()

    config = confighelper.Config(args.configfile)
    print(f"running reconstruction for '{args.configfile}'")
    do_reconstruction(args.num, config)
