// interface between Python and the C++ implementation
// https://docs.microsoft.com/de-de/visualstudio/python/working-with-c-cpp-python-in-visual-studio?view=vs-2019
// https://docs.scipy.org/doc/numpy-1.13.0/user/c-info.how-to-extend.html

#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION

// Make "s#" use Py_ssize_t rather than int.
#define PY_SSIZE_T_CLEAN
#ifdef _DEBUG
    #undef _DEBUG
	#include <Python.h>
	#define _DEBUG
#else
    #include <Python.h>
#endif

#include <stdexcept>
#include <cstdio>
#include <limits>
#include <signal.h>
#include <memory>
// #include "numpy/noprefix.h" includes "no-prefix option" of the types and methods, i.e. intp instead of npy_intp
// defining NPY_NO_PREFIX prior to including "numpy/arrayobject.h" does this as well
#include "numpy/arrayobject.h"

#include "hamiltonians.hpp"


class InterruptExc : public std::exception
{
public:
    InterruptExc() {}
};

void signal_handler(int code)
{
   throw InterruptExc();
}

class PyObjCollection
{
private:
    std::vector<PyObject *> objects;

public:
    /**
     * Validates an PyObject or PyArrayObject.
     * If it fails, a specified error will be returned to Python and all stored pointers will be dereferenced.
     * Otherwise it adds the pointer to the list of stored pointers.
     */
    template <class T>
    bool validate(T *obj, PyObject *error, const char *msg)
    {
        if (!obj)
        {
            PyErr_SetString(error, msg);
            DECREF();
            return false;
        }

        objects.push_back(_PyObject_CAST(obj));
        return true;
    }

    void DECREF()
    {
        for (auto py_obj : objects)
        {
            Py_DECREF(py_obj);
        }
    }
};

template <class... Args>
auto format(const char *format_str, Args&&... args) -> std::unique_ptr<char[]>
{
    int size_s = std::snprintf(nullptr, 0, format_str, args ...);
    if(size_s <= 0){ throw std::runtime_error("Error during formatting."); }
    auto size = static_cast<size_t>(size_s);
    std::unique_ptr<char[]> msg(new char[size]);
    std::sprintf(msg.get(), format_str, args...);
    return msg;
}
// auto convert_to_array(PyObject *pyobj, int dtype, int minDim, int maxDim) -> PyArrayObject *
// {
//     // perform type check
//     if (PyArray_ObjectType(pyobj, dtype) != dtype)
//         return NULL;
    
//     return (PyArrayObject *)PyArray_FROMANY(pyobj, dtype, minDim, maxDim, NPY_ARRAY_IN_ARRAY);
// }

// we don't want the function names to get 'mangled' by the compiler
#ifdef __cplusplus
extern "C"
{
#endif

static auto factored_amplitude(PyObject *self, PyObject *args, PyObject *kwargs) -> PyObject *
{
    // will be const_cast<char**>() below, hopefully PyArg_ParseTupleAndKeywords doesn't try to alter it
    const char *keywords[] = {"c", "tau1", "x_s", "dx", "delta", "max_iter", NULL};
    // define placeholders
    PyObject *pc;
    PyObject *ptau1;
    PyObject *px_s;
    PyObject *pdx;
    double delta;
    usize max_iter = std::numeric_limits<usize>::max();
    // parse arguments
    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "OOOOd|k", const_cast<char**>(keywords), &pc, &ptau1, &px_s, &pdx, &delta, &max_iter))
        return NULL;

    PyObjCollection collection;

    // velocity c
    PyArrayObject *c = (PyArrayObject *)PyArray_FROMANY(pc, NPY_DOUBLE, 2, 3, NPY_ARRAY_IN_ARRAY);
    if (!collection.validate(c, PyExc_ValueError, "c must be a 2D or 3D array of doubles."))
        return NULL;

    const int ndim = PyArray_NDIM(c);

    // traveltime factor
    PyArrayObject *tau1 = (PyArrayObject*)PyArray_FROMANY(ptau1, NPY_DOUBLE, 2, 3, NPY_ARRAY_IN_ARRAY);
    if (!collection.validate(tau1, PyExc_ValueError, "tau1 must be a 2D or 3D array of doubles."))
        return NULL;
    else if (PyArray_NDIM(tau1) != ndim)
    {
        PyErr_SetString(PyExc_ValueError, "The dimension of tau1 must match that of c.");
        collection.DECREF();
        return NULL;
    }

    // check if px_s is a python-numpy-array and if it has the correct size
    // be careful as some loss of information can happen here, e.g. when it is a double-array it will simply be cast to ints
    // signed ints will also silently be cast to unsigned ints and therefore underflow
    PyArrayObject *x_s_ = (PyArrayObject*)PyArray_FROMANY(px_s, NPY_LONG, 1, 1, NPY_ARRAY_IN_ARRAY);
    if (!collection.validate(x_s_, PyExc_ValueError, "x_s must be a 1D array of nonnegative ints."))
        return NULL;
    else if (PyArray_SIZE(x_s_) != ndim)
    {
        PyErr_SetString(PyExc_ValueError, "size of x_s must match the dimension of c.");
        collection.DECREF();
        return NULL;
    }

    // check if pdx is a python-numpy-array and if it has the correct size
    PyArrayObject *dx_ = (PyArrayObject*)PyArray_FROMANY(pdx, NPY_DOUBLE, 1, 1, NPY_ARRAY_IN_ARRAY);
    if (!collection.validate(dx_, PyExc_ValueError, "dx must be a 1D array of doubles."))
        return NULL;
    else if (PyArray_SIZE(dx_) != ndim)
    {
        PyErr_SetString(PyExc_ValueError, "size of dx must match the dimension of c.");
        collection.DECREF();
        return NULL;
    }


    // pointers to data
    double *dx_p = (double *)PyArray_DATA(dx_);
    usize *x_s_p = (usize *)PyArray_DATA(x_s_);

    usize shape[ndim];
    // index version of x_s_
    usize x_s_i = 0;

    usize tmp = 1;
    for (int d = ndim-1; d >= 0; d--)
    {
        shape[d] = PyArray_DIM(c, d);

        if (shape[d] != (usize)PyArray_DIM(tau1, d))
        {
            PyErr_SetString(PyExc_ValueError, format("c and tau1 must have the same shape, but for axis %d the shapes are %ld != %ld.", d, shape[d], PyArray_DIM(tau1, d)).get());
            collection.DECREF();
            return NULL;
        }
        if (x_s_p[d] >= shape[d])
        {
            PyErr_SetString(PyExc_ValueError, format("entries of x_s have to be within the shape of c, but entry %d with value %ld is not in [0, %ld].", d, x_s_p[d], shape[d]-1).get());
            collection.DECREF();
            return NULL;
        }
        if (dx_p[d] <= 0)
        {
            PyErr_SetString(PyExc_ValueError, format("entries of dx must be greater than zero, but entry %d with value %f is not.", d, dx_p[d]).get());
            collection.DECREF();
            return NULL;
        }

        x_s_i += tmp * x_s_p[d];
        tmp *= shape[d];
    }

    // create a new array for the return value
    PyArrayObject *res = (PyArrayObject*)PyArray_ZEROS(ndim, PyArray_DIMS(c), NPY_DOUBLE, 0);
    if (!collection.validate(res, PyExc_MemoryError, "couldn't create result array."))
        return NULL;

    // install custom sigintterrupt handler (CTRL+C)
    auto handler_sigint = signal(SIGINT, signal_handler);
    double error;
    bool success = true;

    try
    {
        if (ndim == 2)
        {
            usize shape_[2];
            double dx_[2];
            for (int i=0; i < 2; i++)
            {
                shape_[i] = shape[i];
                dx_[i] = dx_p[i];
            }
            SweepInfo<2> info{shape_, dx_};
            FactoredAmplitude<2> hamiltonian{info, (double *)PyArray_DATA(c), (double *)PyArray_DATA(tau1)};
            Sweeper<2> sweeper{info, hamiltonian};

            error = sweeper.solve(x_s_i, delta, (double *)PyArray_DATA(res), max_iter);
        }
        else if (ndim == 3)
        {
            usize shape_[3];
            double dx_[3];
            for (int i=0; i < 3; i++)
            {
                shape_[i] = shape[i];
                dx_[i] = dx_p[i];
            }
            SweepInfo<3> info{shape_, dx_};
            FactoredAmplitude<3> hamiltonian{info, (double *)PyArray_DATA(c), (double *)PyArray_DATA(tau1)};
            Sweeper<3> sweeper{info, hamiltonian};

            error = sweeper.solve(x_s_i, delta, (double *)PyArray_DATA(res), max_iter);
        }
        else // can never happen, just to satisfy the compiler
        {
            error = 0;
        }
    }
    catch (const InterruptExc& ex)
    {
        PyErr_SetString(PyExc_KeyboardInterrupt, "");
        collection.DECREF();
        success = false;
    }
    catch (const std::exception& ex)
    {
        // propagate error
        PyErr_SetString(PyExc_RuntimeError, ex.what());
        collection.DECREF();
        success = false;
    }

    // restore original signal handler
    signal(SIGINT, handler_sigint);

    if (!success)
        return NULL;

    return Py_BuildValue("Od", (PyObject *)res, error);
}

#ifdef __cplusplus
}
#endif


// Each entry in this array is a PyMethodDef structure containing
// 1) the Python name,
// 2) the C - function that implements the function,
// 3) flags indicating whether or not keywords are accepted for this function, and
// 4) The docstring for the function.
static PyMethodDef camplitude_methods[] = {
    {"factored_amplitude", (PyCFunction)factored_amplitude, METH_VARARGS | METH_KEYWORDS,
            "factored_amplitude(c, tau1, x_s, dx, delta, max_iter=None)\n--\n\n"
            "    Calculates amplitude factor of the transport equation.\n"
            "\n"
            "    Parameters\n"
            "    ----------\n"
            "    c : ndarray\n"
            "        backgound velocity (used in the calculation for the initial values).\n"
            "    tau1 : ndarray\n"
            "        factor of the traveltime, obtained by solving the factored eikonal equation.\n"
            "    x_s : sequence of ints\n"
            "        Source position as 2D index vector, e.g. ``(0, 0)``.\n"
            "    dx : sequence of doubles\n"
            "        Grid spacing for each dimension, dx > 0.\n"
            "    delta : double\n"
            "        Convergence threshold.\n"
            "    max_iter : optional, int\n"
            "        maximum number of sweeps.\n"
            "\n"
            "    Returns\n"
            "    ----------\n"
            "    a1 : ndarray\n"
            "        numerical solution a1 for the factored transport equation.\n"
            "    residual : float\n"
            "        the residual error norm of the last sweep."
    },

    // Terminate the array with an object containing nulls.
    {NULL, NULL, 0, NULL}
};


static struct PyModuleDef camplitude_module = {
    PyModuleDef_HEAD_INIT,
    "camplitude",                                                                       // Module name to use with Python import statements
    "c++ implementation of Lax-Friedrich sweeping for the factored transport equation", // Module description
    0,
    camplitude_methods                                                                  // Structure that defines the methods of the module
};


PyMODINIT_FUNC PyInit_camplitude(void)
{
    PyObject *m = PyModule_Create(&camplitude_module);
    if (m == NULL)
        return NULL;

    import_array();
    return m;
}
