#include <cmath>
#include <limits>
#include <algorithm>
#include <cassert>
#include <numeric>
#include <functional>
#include <iostream>

#include "sweeping.hpp"


const double INF = std::numeric_limits<double>::infinity();
const double EPS = std::numeric_limits<double>::epsilon();


auto create_ranges(int ndim, const usize *shape) -> std::vector<std::vector<Range>>
{
    // create vector of 2^ndim entries with a vector of ndim ranges each
    auto ranges = std::vector<std::vector<Range>>(pow(2, ndim));

    for (int i=0; i < pow(2, ndim); i++)
    {
        for (int d=0; d < ndim; d++)
        {
            // ranges are indexed from 1 to avoid underflow for negative sweeping directions (see for loops in 'sweep')
            if (((i >> d) & 1) == 0)
            {
                // forward iteration
                ranges[i].push_back({1, shape[d]+1, 1});
            }
            else
            {
                // backward iteration
                ranges[i].push_back({shape[d], 0, -1});
            }
        }
    }

    return ranges;
}

// Note: the returned pointer has to be deleted manually
auto calculate_shifts(int ndim, const usize *shape) -> ssize *
{
    ssize *shift = new ssize[ndim];
    ssize c_shift = 1;

    for (int i = ndim - 1; i >= 0; i--)
    {
        shift[i] = c_shift;
        c_shift *= shape[i];
    }

    return shift;
}

template <int NDIM>
SweepInfo<NDIM>::SweepInfo(const usize (&shape)[NDIM], const double (&dx)[NDIM]):
    shape(shape),
    size(std::accumulate(shape, shape+NDIM, 1, std::multiplies<usize>())),
    dx(dx),
    shift(calculate_shifts(NDIM, shape)),
    sweep_ranges(create_ranges(NDIM, shape))
{}
// explicitly instanciate template function
template SweepInfo<2>::SweepInfo(const usize (&shape)[2], const double (&dx)[2]);
template SweepInfo<3>::SweepInfo(const usize (&shape)[3], const double (&dx)[3]);


// linear extrapolation
template <int NDIM>
void Sweeper<NDIM>::extrapolate(const usize x, int d, double *const res, double &ll, double &l, double &r, double &rr) const
{
    // get index in axis 'd'
    usize x_d = x;
    for (int i=0; i < d; i++)
    {
        x_d = x_d % info.shift[i];
    }
    x_d = x_d / info.shift[d];

    // enough points for stencil
    if (x_d > 1 && x_d < info.shape[d]-2)
    {
        ll = res[x - 2*info.shift[d]];
        l = res[x - info.shift[d]];
        r = res[x + info.shift[d]];
        rr = res[x + 2*info.shift[d]];
    }
    // left boundary
    else if (x_d == 1)
    {
        ll = 2*res[x - info.shift[d]] - res[x];
        l = res[x - info.shift[d]];
        r = res[x + info.shift[d]];
        rr = res[x + 2*info.shift[d]];
    }
    else if (x_d == 0)
    {
        ll = 3*res[x] - 2*res[x + info.shift[d]];
        l = 2*res[x] - res[x + info.shift[d]];
        r = res[x + info.shift[d]];
        rr = res[x + 2*info.shift[d]];
    }
    //right boundary
    else if (x_d == info.shape[d]-2)
    {
        ll = res[x - 2*info.shift[d]];
        l = res[x - info.shift[d]];
        r = res[x + info.shift[d]];
        rr = 2*res[x + info.shift[d]] - res[x];
    }
    else
    {
        ll = res[x - 2*info.shift[d]];
        l = res[x - info.shift[d]];
        r = 2*res[x] - res[x - info.shift[d]];
        rr = 3*res[x] - 2*res[x - info.shift[d]];
    }
}

template<int NDIM>
auto Sweeper<NDIM>::solve_eq(const usize x, double *const res) const -> double
{
    // container
    double grad_u[NDIM];
    double alpha_fac = 0;
    double remainder = 0;

    for (auto d = 0; d < NDIM; d++)
    {
        // ll = res[x-2],   l = res[x-1],  r = res[x+1],    rr = res[x+2]
        double ll, l, r, rr;
        extrapolate(x, d, res, ll, l, r, rr);

        double denom = EPS + pow(r - 2 * res[x] + l, 2);
        double gamma, omega;

        gamma = (EPS + pow(res[x] - 2 * l + ll, 2)) / denom;
        omega = 1 / (1 + 2 * pow(gamma, 2));
        double du_minus = ((1 - omega) * (r - l) + omega * (3 * res[x] - 4 * l + ll)) / (2 * info.dx[d]);

        gamma = (EPS + pow(res[x] - 2 * r + rr, 2)) / denom;
        omega = 1 / (1 + 2 * pow(gamma, 2));
        double du_plus = ((1 - omega) * (r - l) + omega * (-3 * res[x] + 4 * r - rr)) / (2 * info.dx[d]);

        grad_u[d] = (du_minus + du_plus) / 2;
        //remainder += alpha[d] * (res[x.z] / info.dx[d] + (du_plus - du_minus) / 2); -> res[x.z] / info.dx[d] term will be added below
        remainder += alpha[d][x] * (du_plus - du_minus) / 2;
        alpha_fac += alpha[d][x] / info.dx[d];
    }

    auto h = hamiltonian.H(x, res[x], grad_u);
    // f is always 0 in our notation
    //return (1 / alpha_fac) * (- h + remainder);
    return (1 / alpha_fac) * (-h + remainder) + res[x];
}


template<>
auto Sweeper<2>::sweep(const std::vector<Range> &ax_ranges, bool *const fixed, double *const res) const -> double
{
    // norm of subsequent iterates
    double residual = 0;

    // ranges are indexed from 1 to avoid underflow
    for (usize x=ax_ranges[0].start; x != ax_ranges[0].stop; x += ax_ranges[0].step)
    {
        for (usize y=ax_ranges[1].start; y != ax_ranges[1].stop; y += ax_ranges[1].step)
        {
            usize x_i = info.get_index({x-1, y-1});
            if (fixed[x_i])
                continue;

            // calculations
            auto val = solve_eq(x_i, res);

            // residual = residual + fabs(val - res[p.z]); // L1 norm
            if (fabs(val - res[x_i]) > residual)
            {
                residual = fabs(val - res[x_i]); // maximum norm
            }
            // update value
            res[x_i] = val;
        }
    }

    return residual;
}

template<>
auto Sweeper<3>::sweep(const std::vector<Range> &ax_ranges, bool *const fixed, double *const res) const -> double
{
    // norm of subsequent iterates
    double residual = 0;

    // ranges are indexed from 1 to avoid underflow
    for (usize x=ax_ranges[0].start; x != ax_ranges[0].stop; x += ax_ranges[0].step)
    {
        for (usize y=ax_ranges[1].start; y != ax_ranges[1].stop; y += ax_ranges[1].step)
        {
            for (usize z=ax_ranges[2].start; z != ax_ranges[2].stop; z += ax_ranges[2].step)
            {
                usize x_i = info.get_index({x-1, y-1, z-1});
                if (fixed[x_i])
                    continue;

                // calculations
                auto val = solve_eq(x_i, res);

                // residual = residual + fabs(val - res[p.z]); // L1 norm
                if (fabs(val - res[x_i]) > residual)
                {
                    residual = fabs(val - res[x_i]); // maximum norm
                }
                // update value
                res[x_i] = val;
            }
        }
    }

    return residual;
}


template<>
double Sweeper<2>::solve(const usize x0, const double delta, double *const res, const usize max_iter)
{
    hamiltonian.initialize(info, x0, res, alpha);
    for (usize i=0; i < info.size; i++)
    {
        assert(alpha[0][i] > 0);
        assert(alpha[1][i] > 0);
    }
    std::cout << "[camplitude] Artificial viscosities: " << alpha[0][0] << ", " << alpha[1][0] << std::endl;

    auto p = info.get_vector(x0);

    bool *fixed = new bool[info.size];
    for (usize i=0; i < info.size; i++)
        fixed[i] = false;

    // fix values in a 3x3 square around the source
    int k = 1;
    for (int x_off=-(int)std::min<usize>(k, p[0]); x_off <= (int)(info.shape[0] - std::max<usize>(info.shape[0]-k, p[0]+1)); x_off++)
    {
        for (int y_off=-(int)std::min<usize>(k, p[1]); y_off <= (int)(info.shape[1] - std::max<usize>(info.shape[1]-k, p[1]+1)); y_off++)
        {
            fixed[x0 + info.get_index<ssize>({x_off, y_off})] = true;
        }
    }

    delete[] p;

    // Serna, Susana, and Jianliang Qian. "A stopping criterion for higher-order sweeping schemes for static Hamilton-Jacobi equations." Journal of Computational Mathematics (2010): 552-568.
    // -> additional stopping criterion (in our case p=3)
    double trunc = pow(std::max<double>(info.dx[0], info.dx[1]), 3);

    usize n_sweeps = 0;
    double last_residual;
    double residual = INF;
    do
    {
        last_residual = residual;
        residual = sweep(info.sweep_ranges[n_sweeps % info.sweep_ranges.size()], fixed, res);
        n_sweeps++;
    }
    while (((residual > delta) || ((residual + trunc) / last_residual <= 1)) && (n_sweeps < max_iter));

    delete[] fixed;
    std::cout << "[camplitude] Sweeps: " << n_sweeps << std::endl;
    return residual;
}

template<>
double Sweeper<3>::solve(const usize x0, const double delta, double *const res, const usize max_iter)
{
    hamiltonian.initialize(info, x0, res, alpha);
    for (usize i=0; i < info.size; i++)
    {
        assert(alpha[0][i] > 0);
        assert(alpha[1][i] > 0);
        assert(alpha[2][i] > 0);
    }
    std::cout << "[camplitude] Artificial viscosities: " << alpha[0][0] << ", " << alpha[1][0] << ", " << alpha[2][0] << std::endl;

    auto p = info.get_vector(x0);

    bool *fixed = new bool[info.size];
    for (usize i=0; i < info.size; i++)
        fixed[i] = false;
    // fix values in a 3x3 square around the source
    int k = 1;
    for (int x_off=-(int)std::min<usize>(k, p[0]); x_off <= (int)(info.shape[0] - std::max<usize>(info.shape[0]-k, p[0]+1)); x_off++)
    {
        for (int y_off=-(int)std::min<usize>(k, p[1]); y_off <= (int)(info.shape[1] - std::max<usize>(info.shape[1]-k, p[1]+1)); y_off++)
        {
            for (int z_off=-(int)std::min<usize>(k, p[2]); z_off <= (int)(info.shape[2] - std::max<usize>(info.shape[2]-k, p[2]+1)); z_off++)
            {
                fixed[x0 + info.get_index<ssize>({x_off, y_off, z_off})] = true;
            }
        }
    }

    delete[] p;

    // Serna, Susana, and Jianliang Qian. "A stopping criterion for higher-order sweeping schemes for static Hamilton-Jacobi equations." Journal of Computational Mathematics (2010): 552-568.
    // -> additional stopping criterion (in our case p=3)
    double trunc = pow(std::max<double>(std::max<double>(info.dx[0], info.dx[1]), info.dx[2]), 3);

    usize n_sweeps = 0;
    double last_residual;
    double residual = INF;
    do
    {
        last_residual = residual;
        residual = sweep(info.sweep_ranges[n_sweeps % info.sweep_ranges.size()], fixed, res);
        n_sweeps++;
    }
    while (((residual > delta) || ((residual + trunc) / last_residual <= 1)) && (n_sweeps < max_iter));

    delete[] fixed;
    std::cout << "[camplitude] Sweeps: " << n_sweeps << std::endl;
    return residual;
}