#pragma once

#include <stddef.h>
#include <vector>
#include <tuple>
#include "typedefs.hpp"


struct Range
{
    const usize start, stop;
    const char step;
};

/**
 * @brief a point in space
 * 
 * @tparam NDIM number of space dimensions
 */
template <int NDIM, class T=usize>
struct Point
{
    const T p[NDIM];

    auto operator[] (T i) const -> const T &
    {
        return p[i];
    };
};


template <int NDIM>
class SweepInfo
{
public:
    // shape (of the mesh)
    const usize (&shape)[NDIM];
    // total number of grid points
    const usize size;
    // grid spacing (for each axis)
    const double (&dx)[NDIM];
    // the shift to apply in every dimension to get a points neighbor (signed since we sometimes multiply this with -1)
	const ssize *shift;

    // sweeping ranges
    const std::vector<std::vector<Range>> sweep_ranges;

    SweepInfo(const usize (&shape)[NDIM], const double (&dx)[NDIM]);

    ~SweepInfo()
    {
        delete[] shift;
    }

    template <class T=usize>
    inline auto get_index(const Point<NDIM, T> p) const -> const T
    {
        T x_i = 0;
        for (int d=0; d < NDIM; d++)
        {
            x_i += p[d] * shift[d];
        }
        return x_i;
    }

    // Warning: the returned pointer has to be deleted manually with 'delete[]'
    template <class T=usize>
    inline auto get_vector(const T x_i) const -> T *
    {
        T rem = x_i;
    
        T *x = new T[NDIM];
        for (int d = 0; d < NDIM; d++)
        {
            x[d] = rem / shift[d];
            rem = rem % shift[d];
        }

        return x;
    }
};


/**
 * @brief Interface for the Problem at hand
 * 
 * @tparam NDIM number of space dimensions
 */
template <int NDIM>
class Hamiltonian
{
public:
    // initializer function
    virtual void initialize(const SweepInfo<NDIM> &info, const usize x0, double *const res, std::array<std::vector<double>, NDIM> &alpha) = 0;
    // the actual hamiltonian
    virtual double H(const usize x, const double u, const double (&du)[NDIM]) const = 0;
};


template <int NDIM>
class Sweeper
{
private:
    // sweep along a given direction (resp. the ax_ranges)
    auto sweep(const std::vector<Range> &ax_ranges, bool *const fixed, double *const res) const -> double;
    // returns the solution of the equation at a specific point 'x'
    auto solve_eq(const usize x, double *const res) const -> double;
    // extrapolates values on the boundary, otherwise returns inner values
    void extrapolate(const usize x, int d, double *const res, double &ll, double &l, double &r, double &rr) const;

protected:
    // the grid-information
    const SweepInfo<NDIM> &info;
    // the problem's hamiltonian
    Hamiltonian<NDIM> &hamiltonian;
    // artificial viscosity (at each point for every dimension)
    std::array<std::vector<double>, NDIM> alpha;

public:
    Sweeper(const SweepInfo<NDIM> &info, Hamiltonian<NDIM> &hamiltonian):
        info(info),
        hamiltonian(hamiltonian)
    {
        alpha.fill(std::vector<double>(info.size));
    }

    virtual ~Sweeper() {}

    // solve the equation H = 0
    double solve(const usize x0, const double delta, double *const res, const usize max_iter);
};