from __future__ import annotations

import math
import sys
from typing import Iterable, Iterator, Optional, Sequence, Tuple, Union

import numpy as np

from util import arrayutils


class Domain:
    dim: int
    area: np.ndarray

    def __init__(self, *ranges: Tuple[float, float]):
        for i, (m, M) in enumerate(ranges):
            assert m < M, f"[{m}, {M}] is not a valid range for axis {i}."

        self.area = np.array(ranges, dtype=float)
        self.dim = len(ranges)
    
    @property
    def start(self) -> np.ndarray:
        return self.area[:, 0]

    @property
    def end(self) -> np.ndarray:
        return self.area[:, 1]

    @property
    def mid(self) -> Union[np.ndarray, float]:
        return np.sum(self.area, axis=1) / 2
    
    def union(self, other: Domain) -> Domain:
        assert self.dim == other.dim, "Dimensions of the domains don't match."

        area = np.empty_like(self.area)
        area[:, 0] = np.min((self.area[:, 0], other.area[:, 0]), axis=0)
        area[:, 1] = np.max((self.area[:, 1], other.area[:, 1]), axis=0)

        return Domain(*area)
    
    def expand(self, axis: int, amount_left: float, amount_right: float) -> None:
        assert 0 <= axis < self.dim

        if amount_left is None:
            amount_left = 0
        if amount_right is None:
            amount_right = 0

        assert np.argmin(self.area[axis] + (-amount_left, amount_right)) == 0, "Axis range would not be valid."
        self.area[axis] += (-amount_left, amount_right)
    
    def expanded(self, axis: int, amount_left: float, amount_right: float) -> Domain:
        d = Domain(*self.area)
        d.expand(axis, amount_left, amount_right)
        return d
    
    def width(self, axis):
        return self.area[axis, 1] - self.area[axis, 0]

    def __getitem__(self, key: Union[int, Tuple[int, int]]) -> Union[Tuple[float], float]:
        return self.area[key]

    def __iter__(self) -> Iterator[Tuple[float, float]]:
        return iter(self.area)
    
    def __repr__(self) -> str:
        return f"Domain({self})"

    def __str__(self) -> str:
        return " x ".join((f"[{start}, {end}]" for start, end in self.area))

class Mesh:
    """
    A simple multi-dimensional cartesian mesh
    """
    domain: Domain
    dx: np.ndarray
    n: Tuple[int, ...]
    X: np.ndarray

    def __init__(self, domain: Domain, n: Union[int, Iterable[Union[int, None]]]):
        if isinstance(n, int):
            n = [n] + [None] * (domain.dim-1)
        else:
            n = list(n)

        assert domain.dim == len(n), "Domain-dimension ({}) does not match dimension of n ({})".format(domain.dim, len(n))

        # look for the first entry in 'n' that is not 'None' and use it's spacing as reference for other dimensions
        ref_dx = None
        for i, nk in enumerate(n):
            if nk is not None:
                ref_dx = domain.width(i) / (nk - 1)
                break
        else:
            raise Exception("n can't consist of only 'None's")


        x = [None] * domain.dim
        self.dx = np.ones(domain.dim)

        for i, nk in enumerate(n):
            n[i] = nk or int(domain.width(i) / ref_dx) + 1
            x[i], self.dx[i] = np.linspace(*domain[i], num=n[i], retstep=True)

        self.domain = domain
        self.n = tuple(n)
        self.X = np.array(np.meshgrid(*x, indexing="ij"))

    @classmethod
    def from_dx(cls, domain: Domain, dx: Union[float, Sequence[Union[float, None]]]) -> Mesh:
        if isinstance(dx, float):
            dx = [dx] + [None] * (domain.dim-1)
        else:
            dx = list(dx)
        
        assert domain.dim == len(dx), "area-dimension ({}) does not match dimension of dx ({})".format(domain.dim, len(dx))

        n = [None] * domain.dim
        for i in range(domain.dim):
            if dx[i] is not None:
                n[i] = int(domain.width(i) / dx[i]) + 1
        
        return Mesh(domain, n)
    
    @property
    def dim(self) -> int:
        return self.domain.dim
    
    def expanded(self, axis: int, left: Optional[float], right: Optional[float]) -> Mesh:
        """
        Creates a new Mesh object with an axis expanded to the left and/or right while maintaining the spacing.
        The new end points may not match exactly the new domain defined by subtracting 'left' and adding 'right' to the current one,
        depending on the spacing.
        """
        ll, rr = self.domain[axis]
        nn = self.n[axis]

        if left is not None:
            assert left > 0, "expand value must be greater than zero"
            n = math.ceil(left / self.dx[axis])
            ll -= n * self.dx[axis]
            nn += n

        if right is not None:
            assert right > 0, "expand value must be greater than zero"
            n = math.ceil(right / self.dx[axis])
            rr += n * self.dx[axis]
            nn += n

        area = list(self.domain.area)
        area[axis] = ll, rr
        n = list(self.n)
        n[axis] = nn
        return Mesh(Domain(*area), n)
    
    def get_axis(self, axis: int) -> Sequence[float]:
        ind = arrayutils.index_tpl(self.dim, 0, axis, slice(None))
        return self.X[axis][tuple(ind)]

    def index_to_pos(self, x: Union[Sequence, np.ndarray]) -> np.ndarray:
        return self.domain.start + x * self.dx

    def pos_to_index(self, x: Union[Sequence, np.ndarray]) -> np.ndarray:
        return ((x - self.domain.start + sys.float_info.epsilon) / self.dx).astype(int)
    
    def axis_to_index(self, x: float, axis: int=0, floor: bool=True) -> int:
        res = (x - self.domain[axis, 0] + sys.float_info.epsilon) / self.dx[axis]
        return int(res) if floor else math.ceil(res)

    def __str__(self) -> str:            
        return f"""Mesh:
    domain: {self.domain}
    n = {self.n}
    dx = {tuple(self.dx)}"""

class Linspace:
    dx: float

    def __init__(self, start: float, end: float, n: int, **kwargs):
        assert start < end, f"[{start}, {end}] is not a valid range"
        assert n >= 2, "n must be at least 2"
        self._x, self.dx = np.linspace(start, end, n, retstep=True, **kwargs)
    
    @classmethod
    def from_dx(cls, start: float, end: float, dx: float) -> Linspace:
        n = int((end - start) / dx) + 1
        return Linspace(start, end, n)

    def expanded(self, left: Optional[float], right: Optional[float]) -> Linspace:
        """
        Creates a new Linspace object expanded to the left and/or right while maintaining the spacing.
        The new end points may not match exactly the new domain defined by subtracting 'left' and adding 'right' to the current one,
        depending on the spacing.
        """
        ll, rr, nn = self._x[0], self._x[-1], len(self._x)

        if left is not None:
            assert left > 0, "expand value must be greater than zero"
            n = math.ceil(left / self.dx)
            ll -= n * self.dx
            nn += n

        if right is not None:
            assert right > 0, "expand value must be greater than zero"
            n = math.ceil(right / self.dx)
            rr += n * self.dx
            nn += n

        return Linspace(ll, rr, nn)
    
    @property
    def mid(self) -> float:
        return (self._x[0] + self._x[-1]) / 2
    
    def get_index(self, x, floor: bool=True) -> int:
        res = (x - self._x[0] + sys.float_info.epsilon) / self.dx
        return int(res) if floor else math.ceil(res)
    
    def __getitem__(self, index):
        return self._x[index]
    
    def __iter__(self):
        return iter(self._x)
    
    def __len__(self):
        return len(self._x)
    
    def __gt__(self, other):
        if isinstance(other, Linspace):
            return self._x > other._x
        else:
            return self._x > other
    
    def __ge__(self, other):
        if isinstance(other, Linspace):
            return self._x >= other._x
        else:
            return self._x >= other
    
    def __lt__(self, other):
        if isinstance(other, Linspace):
            return self._x < other._x
        else:
            return self._x < other
    
    def __le__(self, other):
        if isinstance(other, Linspace):
            return self._x <= other._x
        else:
            return self._x <= other
    
    def __eq__(self, other):
        if isinstance(other, Linspace):
            return self._x == other._x
        else:
            return self._x == other
    
    def __str__(self):
        return f"[{self._x[0]}, {self._x[-1]}],\tn = {len(self._x)},\tdx = {self.dx}"

if __name__ == "__main__":
    l1 = Linspace(-5, 5, 26)
    l2 = l1.expanded(5.2, 5.2)
    print(l1._x)
    print(l2._x)
