from __future__ import annotations
from typing import List, Sequence

import numpy as np
import scipy.sparse as sp

from util import stringutils
from mesh import Linspace, Mesh


def __binary_search(keys, key):
    """ returns:
        -1, if key is smaller then the smallest key
        len(self._keys) - 1, if key is bigger than or equal to the biggest key
        l, where for the left index l: keys[l] < key < keys[l+1]
    """
    if key < keys[0]:
        return -1
    elif key > keys[-1]:
        return len(keys) - 1

    l, r = 0, len(keys)
    while l + 1 < r:
        m = l + (r - l) // 2
        if key == keys[m]:
            return m
        elif key < keys[m]:
            r = m
        else:
            l = m
    return l

def __get_interpolated_kernel(y_ref, kernels_ref, y):
    y_i_ref = __binary_search(y_ref, y)
    if y_i_ref == -1:
        return kernels_ref[0]
    elif y_i_ref == len(y_ref) - 1:
        return kernels_ref[-1]
    else:
        gamma = (y_ref[y_i_ref+1] - y) / (y_ref[y_i_ref+1] - y_ref[y_i_ref])
        return gamma * kernels_ref[y_i_ref] + (1 - gamma) * kernels_ref[y_i_ref+1]

# import cv2
# def _scale_kernel(K, tar_shape):
#     return cv2.resize(K.T, tar_shape, interpolation=cv2.INTER_LANCZOS4).T

def scalar_prod(p_mesh: Mesh, s: Linspace, t: Sequence[float], y_ref: List[float], kernels_ref: List[np.ndarray], g: np.ndarray):
    """
    calculates the L2-scalar product of Kernels in the p_mesh with the (cutoff) data.

    Parameters
    ----------
        p_mesh : Mesh
            the domain mesh of p.
        s : Linspace
            offsets
        t : sequence of floats
            time points
        y_ref : list of floats
            reference kernel y-coordinates
        kernels_ref : list of NumPy-arrays
            reference kernels matching the y-coordinates 'y_ref'
        g : np.ndarray
            right-hand-side data, cutoff radon transform (see @rhs.py)

    Returns
    ----------
        reconstruction : np.ndarray
            the reconstruction on the p_mesh
    """
    assert len(y_ref) == len(kernels_ref), "number of reference y-coordinates and kernels do not match."
    print("[kernel.py] calculating L2 scalar product")

    mid = (kernels_ref[0].shape[0] - 1) // 2

    interpolate = False
    # TODO: if kernels_ref.shape[0] >> g.shape[0]: maybe use more than just 2 points for interpolation?
    # or we could loose information
    # finer kernels than data
    if len(s) > g.shape[0]:
        points = np.linspace(0, len(s)-1, g.shape[0])
        # lower bound indices
        left = points.astype(int)
        # calculate neighboring indices and ensure they are not outside of arr.shape
        right = np.minimum(left + 1, len(s)-1 - 1)
        # weights for the two neighboring axis
        factor = (points - left)[:, np.newaxis]
        interpolate = True

    res = np.empty(p_mesh.n)
    bar = stringutils.ProgressBar(60, p_mesh.n[0] * p_mesh.n[1])
    # TODO: better
    t_dx = t[1] - t[0]

    for y in range(p_mesh.n[1]):
        p_y = p_mesh.X[1][0, y]
        # get the interpolated reference for the y-coordinate
        K_ref = __get_interpolated_kernel(y_ref, kernels_ref, p_y)

        for x in range(p_mesh.n[0]):
            p_x = p_mesh.X[0][x, 0]
            # cut out the correct slice for the x-coordinate
            x_i = s.get_index(p_x)

            # from the middle of K_ref, which is based on the reference mesh 'mesh "+" p_mesh',
            # we need 'x_i' points to the left and 'len(s) - x_i' to the right
            K = K_ref[mid - x_i: mid + (len(s) - x_i), :]

            # interpolate in x-direction
            if interpolate:
                if sp.issparse(K):
                    K = K[left, :].multiply(1 - factor) + K[right, :].multiply(factor)
                else:
                    K = (1 - factor) * K[left, :] + factor * K[right, :]
                # K = _scale_kernel(K, g.shape)

            # L2-scalar product of cuttoff-data and reconstruction kernel
            # TODO: for this t must also be linspace!
            if sp.issparse(K):
                res[x, y] = s.dx * t_dx * np.sum(K.multiply(g))
            else:
                res[x, y] = s.dx * t_dx * np.sum(K * g)
            bar.tick()

    return res


if __name__ == "__main__":
    from mesh import Domain

    s = Linspace(-10, 10, 201)
    t = Linspace(0, 10, 201)

    p_mesh = Mesh(Domain((-2.5, 5), (0, 7)), 51)
    p_domain = p_mesh.domain

    max_p_axis = max(s.mid - p_domain[0, 0], p_domain[0, 1] - s.mid)
    # try to account for rounding errors later on
    if (max_p_axis / s.dx) % 1 != 0:
        max_p_axis += s.dx

    s_ref = s.expanded(max_p_axis, max_p_axis)
    y_ref = list(range(7))

    kernels_ref = [i * np.ones((len(s_ref), len(t))) for i in y_ref]
    g = np.ones((101, len(t)))

    scalar_prod(p_mesh, s, t, y_ref, kernels_ref, g)
