"""
This module is all about the right-hand side of the equation Fn = y.
"""
from multiprocessing import Pool
from typing import Callable, List, Sequence, Tuple, Union
from matplotlib import animation

import matplotlib.pyplot as plt
import numpy as np
from numpy.lib import math
from scipy import interpolate

import GRT, kernel
from GRT import GRTData
from mesh import Linspace, Mesh
from util import mphelper, stringutils

try:
    import pysit
except:
    print("[rhs.py] Warning - pysit not installed, which is required for ForwardSolver to work.")


# --------------------------------------------------
# Cutoff functions
# --------------------------------------------------
def __uu(r: np.ndarray) -> np.ndarray:
    res = np.zeros_like(r)
    res[r > 0] = np.exp(-1 / r[r > 0])
    return res

def __cutoff_left(v: np.ndarray, start: float, end: float) -> np.ndarray:
    return __uu(v - start) / (__uu(v - start) + __uu(end - v))
    # original version
    # return __uu(v - start) / (__uu(v - start) + __uu((start + end) / 2 - v))

def __cutoff_right(v: np.ndarray, start: float, end: float) -> np.ndarray:
    return __uu(end - v) / (__uu(end - v) + __uu(v - start))
    # original version
    # return __uu(end - v) / (__uu(end - v) + __uu(v - (start + end) / 2))

def cut_off_axis(v: Sequence[float], lower: Union[Tuple[float, float], None], upper: Union[Tuple[float, float], None]) -> np.ndarray:
    """
    Smooth cutoff (to zero) for the array v.
    """
    if not (lower is None or upper is None):
        assert lower[1] <= upper[0], f"lower cutoff must be smaller than higher one. min-end:{lower[1]} - max-start:{upper[0]}"

    v = np.asarray(v)
    res = np.ones_like(v)

    if lower is not None:
        assert lower[0] <= lower[1], f"cutoffs must be increasing. start:{lower[0]} - end:{lower[1]}"
        res[v <= lower[0]] = 0
        res[(v > lower[0]) & (v < lower[1])] = __cutoff_left(v[(v > lower[0]) & (v < lower[1])], *lower)

    if upper is not None:
        assert upper[0] <= upper[1], f"cutoffs must be increasing. start:{upper[0]} - end:{upper[1]}"
        res[v >= upper[1]] = 0
        res[(v > upper[0]) & (v < upper[1])] = __cutoff_right(v[(v > upper[0]) & (v < upper[1])], *upper)

    return res

def cut_off(s: np.ndarray, t: np.ndarray, S_lower: Union[Tuple[float, float], None], S_upper: Union[Tuple[float, float], None],
            T_lower: Union[Tuple[float, float], None], T_upper: Union[Tuple[float, float], None]) -> np.ndarray:
    """
    Smooth cutoff (to zero) for the 2D-array s x t.
    """
    return np.outer(cut_off_axis(s, S_lower, S_upper), cut_off_axis(t, T_lower, T_upper))

def cut_off_centered(s: np.ndarray, t: np.ndarray, S: Union[Tuple[float, float], None], T_lower: Union[Tuple[float, float], None], T_upper: Union[Tuple[float, float], None]) -> np.ndarray:
    """
    s-centered cutoff
    """
    return cut_off(s, t, (-S[1], -S[0]), S, T_lower, T_upper)


# --------------------------------------------------
# Radon transform for a phantom n-function
# --------------------------------------------------
def __init_worker(dx, phi_info, integrand_factor_info, y_min, s, n_func, res_info):
    # use global variables to save shared info
    global __grt_data, __y_min, __s, __n_func, __res
    __grt_data = GRTData(dx, mphelper.np_from_shared(phi_info), mphelper.np_from_shared(integrand_factor_info))
    __y_min, __s, __n_func, __res = y_min, s, n_func, mphelper.np_from_shared(res_info)

def __worker_func(t: float, t_i: int, *n_args):
    # write result row into shared array
    GRT.F_tilde(__grt_data, __y_min, __s, t, __n_func, *n_args, res=__res[:, t_i])

def calculate_rt(mesh: Mesh, grt_data: GRTData, s: Linspace, t: Sequence[float], n_func: Callable[[Sequence[Tuple[float, float]]], Sequence[float]]) -> np.ndarray:
    """
    Calculates the radon transform 'F*n' for a given phantom 'n'.

    Parameters
    ----------
        mesh : Mesh
            the domain mesh of c.
        grt_data : GRTData
            return value of 'GRT.calculate_F_data'
        s : Linspace
            offsets for the kernel
        t : sequence of floats
            list of times for which the contours are calculated (for example a suitable 'np.linspace' array)
        n_func : function
            phantom n function. Should accept one parameter being the points at which it will be evaluated
            and return a list of those values.

    Returns
    ----------
        Fn : ndarray
            result-array
    """
    # prepare result array
    res, res_info = mphelper.create_shared((len(s), len(t)), float)
    shared_info = kernel._create_shared_info(grt_data, mesh.domain[1, 0], np.asarray(s), n_func, res_info)

    with Pool(initializer=__init_worker, initargs=shared_info) as pool:
        print("[rhs.py] Calculating radon transform using {} cores.".format(pool._processes))
        bar = stringutils.ProgressBar(60, len(t))

        futures = []
        for t_i in range(len(t)):
            fut = pool.apply_async(__worker_func, (t[t_i], t_i))
            futures.append(fut)

        for fut in mphelper.as_completed(futures):
            fut.get() # raise Exceptions that occured
            bar.tick()

        pool.close()
        pool.join()

    return res

def add_ball(arr: np.ndarray, x: Sequence[Tuple[float, float]], m: Tuple[float, float], r: float, value: float, ord=2) -> None:
    """
    Adds 'value' to all points in 'arr' contained in a ball around 'm' with radius 'r'.
    """
    dist = np.linalg.norm(x - m, ord=ord, axis=-1)
    arr[dist < r] += value

def add_floor(arr: np.ndarray, x: Sequence[Tuple[float, float]], k: float, value: float) -> None:
    """
    Adds 'value' to all points in 'arr' with second axis greater than 'k'.
    """
    arr[x[:, 1] >= k] += value

def add_sin(arr: np.ndarray, x: Sequence[Tuple[float, float]], y: float, a: float, p: float, o: float, value: float) -> None:
    """
    Adds 'value' to all points in 'arr' with second axis greater than 'y + a sin(p x_2 + o)'.
    """
    s = y + a * np.sin(p * x[:, 0] + o)
    arr[x[:, 1] >= s] += value

def add_rect(arr: np.ndarray, x: Sequence[Tuple[float, float]], m: Tuple[float, float], r: Tuple[float, float], value: float) -> None:
    """
    Adds 'value' to all points in 'arr' contained in a rectangle around 'm' with edge-length '2r'.
    """
    dist = np.abs(x - m)
    arr[np.all(dist < r, axis=-1)] += value

def add_rot_rect(arr: np.ndarray, x: Sequence[Tuple[float, float]], m: Tuple[float, float], r: Tuple[float, float], angle: float, value: float) -> None:
    """
    Adds 'value' to all points in 'arr' contained in a rectangle rotated by 'angle' around 'm' with edge-length '2r'.
    """
    r = np.array(r)
    # project points to AABB at the origin
    proj = (np.array([[np.cos(angle), np.sin(angle)],
                     [-np.sin(angle), np.cos(angle)]]) @ (x - m).T).T
    arr[np.all(np.abs(proj < r), axis=1)] += value

def plot_n_func(mesh, n_func):
    points = np.vstack((mesh.X[0].ravel(), mesh.X[1].ravel())).T
    res = n_func(points)
    res = res.reshape(mesh.n)

    plt.pcolormesh(*mesh.X, res, cmap="Greys", rasterized=True, shading="auto")


# --------------------------------------------------
# Real data
# --------------------------------------------------
def generate_seismograms(shot, solver, base_model):
    pysit.generate_shot_data_time(shot, solver, base_model)
    return shot.receivers.data, shot.receivers.ts

class ForwardSolver:
    def __init__(self, x_mesh: Mesh, pmlx: Tuple[float, float]=(0.1, 0.1), pmly: Tuple[float, float]=(0.1, 0.1)) -> None:
        minx, miny = x_mesh.domain.start
        maxx, maxy = x_mesh.domain.end

        # TODO: adapt PML to mesh size automatically
        pmlx = pysit.PML(pmlx[0], 100), pysit.PML(pmlx[1], 100)
        pmly = pysit.PML(pmly[0], 100), pysit.PML(pmly[1], 100)
        configx = minx, maxx, *pmlx
        configy = miny, maxy, *pmly

        domain = pysit.RectangularDomain(configx, configy)
        self.mesh = pysit.CartesianMesh(domain, *x_mesh.n)

    def generate_shots(self, s_list: Sequence[float], alpha: float) -> Sequence["pysit.Shot"]:
        """
        Generates the shots (source and receiver informations) for the common offset geometry.
        We use a GaussianPulse to approximate a delta distribution source.
        """
        wavelet = pysit.GaussianPulse(200.0, t_shift=0.2)

        y_min = self.mesh.domain.z.lbound

        shots = []
        for s in s_list:
            # create shot
            source = pysit.PointSource(self.mesh, (s-alpha, y_min), wavelet)
            receiver = pysit.PointReceiver(self.mesh, (s+alpha, y_min))
            shots.append(pysit.Shot(source, receiver))
        
        return shots

    def solve(self, shots: Sequence["pysit.Shot"], t_range: Tuple[float, float], v: np.ndarray, output_wavefields: bool=False) -> Tuple[np.ndarray, Sequence[float], List[List[np.ndarray]]]:
        """
        Solves the wave equation
        1/v^2 d^2u/dt^2 - Lapl(u) = f
        with the given parameters

        Parameters
        ----------
            shots : sequence of pysit.Shot
                the shots (source and receiver informations) used. Can be obtained by calling 'generate_shots'
            t_range : 2-Tuple of floats
                the time interval in which to solve the equation
            v : numpy-array
                the velocity model used
            output_wavefields : bool (optional)
                if True, a list of wavefields (np.ndarray) for each shot will be generated.

        Returns
        ----------
            seismograms : numpy-array
                the seismograms for each shot in a single (s-t)-array
            t : sequence of float
                the actual time-steps obtained from the solver
            wavefields: sequence of sequence of np.ndarray
                List of wavefields (np.ndarray) for each shot. Each element is a list of wavefields for each time step.
                Only returned, when output_wavefields is True.
        """
        assert len(shots) > 0, "at least one shot is needed"
        v = v.reshape(-1, 1)
        t_shift = shots[0].sources.w.t_shift #- 0.5 / shots[0].sources.w.peak_frequency

        # define the solver
        solver = pysit.ConstantDensityAcousticWave(
            self.mesh,
            formulation = "scalar",
            kernel_implementation = "cpp",
            spatial_accuracy_order = 6, #8
            trange = (t_range[0], t_range[1] + t_shift), # modify t-range to compensate for time-shift
            # cfl_safety=0.1 # this essentially modifies 'dt' (default is cfl=1/6)
        )

        base_model = solver.ModelParameters(self.mesh, {'C': v})

        t = None
        t_i = None

        wavefields = []
        for shot in shots:
            wavefields_shot = [] if output_wavefields else None
            pysit.generate_shot_data_time(shot, solver, base_model, wavefields=wavefields_shot)
            
            # ts is the same for all receivers/shots
            if t is None:
                t = shot.receivers.ts
                # compensate for the time shift of the gaussian pulse
                t_i = np.argmax(t >= t_shift)
                t = t[t_i:] - t_shift

            if output_wavefields:
                wavefields.append(wavefields_shot[t_i:])

        # collect seismograms in a single numpy array
        seismograms = np.empty((len(shots), len(shots[0].receivers.data) - t_i))
        for i, shot in enumerate(shots):
            seismograms[i] = shot.receivers.data[t_i:, 0] # data has shape (len(t), 1)

        if output_wavefields:
            return seismograms, t, wavefields
        else:
            return seismograms, t

def animate(wavefields: Sequence[np.ndarray], mesh: "pysit.CartesianMesh", t: Sequence[float], speed=1.0, scale=None, show=True, repeat=True, **kwargs):
    # fig = plt.figure()
    fig = plt.gcf()
    plt.clf()
    ax = plt.gca()

    im = pysit.vis.plot(wavefields[0], mesh, axis=ax, **kwargs)
    # ax.set_aspect("equal")
    cbar = plt.colorbar()
    title = plt.title('{0:4}/{1}'.format(0,len(wavefields)-1))

    if scale == "fixed":
        pltclim = 1.1*min([d.min() for d in wavefields]), 1.1*max([d.max() for d in wavefields])
    else:
        pltclim = None

    def _animate(i, im, data, mesh, title, cbar, pltclim):
        im = pysit.vis.plot(data[i], mesh, axis=ax, update=im, **kwargs)

        if pltclim is None:
            clim = data[i].min(), data[i].max()
            im.set_clim(clim)
        else:
            im.set_clim(pltclim)

        title.set_text(f"{t[i]:.2f}s - frame {i}/{len(data)-1}")
        plt.draw()
        return (im, )

    data_fps = len(wavefields) / (t[-1] - t[0]) * speed

    # a lot of frames for a single second -> skip some
    if data_fps > 23:
        step = math.ceil(data_fps / 23)
        data_fps /= step
    else:
        step = 1

    interval = max(int(1000 // data_fps), 1)
    print(f"data_fps: {data_fps}, step: {step}, interval: {interval}")
    

    _animate_args = (im, wavefields, mesh, title, cbar, pltclim)
    anim = animation.FuncAnimation(fig, _animate, fargs=_animate_args, #init_func=init,
            interval=interval, frames=range(0, len(wavefields), step), blit=True, repeat=repeat)

    if show:
        plt.show()

    return anim

def interpolate_seismograms(t_a: Sequence[float], seismograms_a: np.ndarray, t_b: Sequence[float], seismograms_b: np.ndarray) -> Tuple[Sequence[float], np.ndarray, np.ndarray]:
    """
    Interpolates two (time, seismograms)-pairs to match the time axis.
    The seismogram arrays are assumed to have the second axis as time axis
    """
    # a is bigger -> interpolate it with t-values from b
    if seismograms_a.shape[1] > seismograms_b.shape[1]:
        interp = interpolate.interp1d(t_a, seismograms_a, axis=1, fill_value="extrapolate", assume_sorted=True)
        return t_b, interp(t_b), seismograms_b
    # b is bigger -> interpolate it with t-values from a
    elif seismograms_a.shape[1] < seismograms_b.shape[1]:
        interp = interpolate.interp1d(t_b, seismograms_b, axis=1, fill_value="extrapolate", assume_sorted=True)
        return t_a, seismograms_a, interp(t_a)
    # same shape -> interpolate nothing and use t from a
    else:
        return t_a, seismograms_a, seismograms_b

def fit_seismograms(t: Sequence[float], seismograms: np.ndarray, seismograms_bg: np.ndarray, search_length: float=0.2):
    """
    Fits 'seismograms' to 'seismograms_bg' by asjusting the amplitude of the initial wave in 'seismograms' to the one in 'seismograms_bg'.

    Parameters
    ----------
        t : sequence of floats
            time steps for both seismograms
        seismograms : numpy-array
            seismograms for the real velocity
        seismograms_bg : numpy-array
            seismograms for the background velocity
        search_length : float (optional)
            the width (in percent of 't') of the search area in 'seismograms'

    Returns
    ----------
        seismograms : numpy-array
            the scaled seismograms
    """
    # find the index of the biggest initial wave amplitude for each offset
    ind = np.argmax(seismograms_bg, axis=1)
    # convert 'search_length' to actual corresponding time interval
    l = 0.5 * search_length * (t[-1] - t[0])

    m_val = 0
    for i, mid in enumerate(ind):
        # find the biggest amplitude in 'seismograms' inside the search interval
        val = np.max(seismograms[i, (t >= t[mid] - l) & (t <= t[mid] + l)])
        if val > m_val:
            m_val = val
            # j = i # debug

    # debug
    # a = 1.05 * max(np.max(np.abs(seismograms)), np.max(np.abs(seismograms_bg)))
    # for i in range(seismograms.shape[0]):
    #     plt.plot(t, i * a + seismograms_bg[i], color="black")
    #     plt.plot(t, i * a + seismograms[i], color="red")

    # plt.vlines(t[ind[j]], (j-0.2) * a, (j+1) * a, color="black")
    # plt.vlines(t[ind[j]], -0.8*a, (j-0.2) * a, color="black", linestyle="--")
    # k = np.where(seismograms[j] == m_val)
    # plt.vlines(t[k], (j-0.2) * a, (j+1) * a, color="red")

    # mask = (t >= t[ind[j]] - l) & (t <= t[ind[j]] + l)
    # plt.plot(t, -a * mask, color="grey", alpha=0.5)
    # plt.show()

    if m_val == 0:
        return seismograms
    else:
        return np.max(seismograms_bg[:, ind]) / m_val * seismograms


if __name__ == "__main__":
    t = np.linspace(0, 1, 1001)

    def gauss(x, mu, sigm):
        return np.exp(-(x - mu)**2 / (2 * sigm**2))

    shape = 5, len(t)
    s = np.empty(shape)
    s_bg = np.empty(shape)
    for i in range(shape[0]):
        o = abs(i - (shape[0]-1) // 2) / shape[0]
        o *= 0.2

        s[i] = (i+1) * gauss(t, 0.2 + o, 0.02) + gauss(t, 0.7+o, 0.01)
        s_bg[i] = gauss(t, 0.2 + o + i*0.01, 0.02)

    s2 = fit_seismograms(t, s, s_bg)

    a = 1.05 * max(np.max(s), np.max(s_bg))
    for i in range(shape[0]):
        plt.plot(t, i * a + s[i], color="red")
        plt.plot(t, i * a + s_bg[i], color="black")
        plt.plot(t, i * a + s2[i], color="green", linestyle="--")
    plt.show()

    pulse = pysit.GaussianPulse(500.0, t_shift=0.2)
    y = pulse(t)
    plt.plot(t, y)
    plt.vlines(0.2, 0, 1, color="red")
    plt.show()
