import sys
import numpy as np
from typing import Any, Callable, Iterable, Sequence, Tuple, TypeVar, Union

# class SupportsProduct(Protocol):
#     def __mul__(self, __other: Any) -> bool: ...

_T = TypeVar("_T", float, int, np.ndarray)


def dot(a: Iterable[_T], b: Iterable[_T]) -> _T:
    """
    calculates the dot product of two vectors
    """
    a_it, b_it = iter(a), iter(b)
    res: _T = next(a_it) * next(b_it)
    for x, y in zip(a_it, b_it):
        res += x * y

    return res

def interp_index(points: np.ndarray, array: np.ndarray) -> np.ndarray:
    """
    (bi-)linearly interpolates float indices 'points' of a 2D array
    """
    assert array.ndim == points.shape[1], "ndim of 'arr' and 2nd shape of 'points' does not match!"
    shape = array.shape

    # calculate lower bound indices
    rounded = points.astype(int)
    x, y = rounded[:, 0], rounded[:, 1]
    # calculate the weights for the four raster values
    factor = points - rounded
    factor_x, factor_y = factor[:, 0], factor[:, 1]
    # calculate neighboring indices and ensure they are not outside of arr.shape
    x_right = np.minimum(x + 1, shape[0] - 1)
    y_right = np.minimum(y + 1, shape[1] - 1)

    # interpolate in x-direction
    arr_interp_x1 = (1 - factor_x) * array[x, y] + factor_x * array[x_right, y]
    arr_interp_x2 = (1 - factor_x) * array[x, y_right] + factor_x * array[x_right, y_right]

    # interpolate the results in y-direction
    return (1 - factor_y) * arr_interp_x1 + factor_y * arr_interp_x2

'''
def interps_index(points: np.ndarray, *arrays: np.ndarray) -> Sequence[np.ndarray]:
    """
    (bi-)linearly interpolates float indices 'points' of multiple 2D arrays
    """
    assert len(arrays) > 1, "at least two 2D arrays should be provided. For a single array, use 'interp_index'"
    assert arrays[0].ndim == points.shape[1], "ndim of 'arr' and 2nd shape of 'points' does not match!"
    shape = arrays[0].shape

    # calculate lower bound indices
    rounded = points.astype(int)
    x, y = rounded[:, 0], rounded[:, 1]
    # calculate the weights for the four raster values
    factor = points - rounded
    factor_x, factor_y = factor[:, 0], factor[:, 1]
    # calculate neighboring indices and ensure they are not outside of arr.shape
    x_right = np.minimum(x + 1, shape[0] - 1)
    y_right = np.minimum(y + 1, shape[1] - 1)

    res = []
    for arr in arrays:
        assert arr.shape == shape, "the shape of one of the arrays does not match the first one. {} != {}".format(shape, arr.shape)
        # interpolate in x-direction
        arr_interp_x1 = (1 - factor_x) * arr[x, y] + factor_x * arr[x_right, y]
        arr_interp_x2 = (1 - factor_x) * arr[x, y_right] + factor_x * arr[x_right, y_right]

        # interpolate the results in y-direction
        res.append((1 - factor_y) * arr_interp_x1 + factor_y * arr_interp_x2)
    
    return res
'''

def get_centered_slice(arr: np.ndarray, x_index: int, width: int=None) -> np.ndarray:
    # if width=None use the maximum possible width
    if width is None:
        width = 2 * min(x_index + 1, arr.shape[0] - x_index) - 1

    left = x_index - (width-1) // 2
    if left < 0 or left + width > arr.shape[0]:
        raise IndexError("Can't slice array of size {} with index starting at {} and width {}".format(arr.shape[0], left, width))

    return arr[left:left+width, :]

def index_tpl(length: int, fill_value: Any, i: int, value: Any) -> Tuple:
    """
    Creates a tuple of given length with 'value' at index 'i' and (references to) 'fill_value' otherwise.
    """
    if i < 0:
        i = length + i
    assert 0 <= i < length, f"index not in range. i={i}, length={length}"

    arr = [fill_value] * length
    arr[i] = value
    return tuple(arr)

def integral_function(x: Sequence[float], f: np.ndarray, axis: int=0):
    """
    Approximates the integral function 'y(r) = int_0^r f(s) ds' for every r in x using the trapezoidal rule.

    Parameters
    ----------
        x : sequence of floats
            the upper bounds for the integral function
        f : numpy-array
            the function to integrate over
        axis : int (optional)
            the axis of f to integrate, default 0

    Returns
    ----------
        integral : numpy-array
            the integral function for each value in x
    """
    assert 0 <= axis < f.ndim, f"f does not have enough dimensions f.ndim={f.ndim}, axis={axis}"
    assert f.shape[axis] == len(x), f"x-values and chosen axis of the function must have the same length. len(x)={len(x)} != {f.shape[axis]} = len(f[axis])"

    idx_left = index_tpl(f.ndim, slice(None), axis, slice(0, -1))
    idx_right = index_tpl(f.ndim, slice(None), axis, slice(1, f.shape[axis]))
    idx = index_tpl(f.ndim, 1, axis, -1)

    x = np.reshape(x[1:] - x[:-1], idx)
    # interpolate using trapezoidal rule
    # y[0] = 0
    # y[j] = y[j-1] + 0.5 * (f[j] + f[j-1]) * (x[j] - x[j-1])
    y = np.zeros_like(f)
    # 0.5 * (f[j] + f[j-1]) * (x[j] - x[j-1])
    y[idx_right] = 0.5 * (f[idx_right] + f[idx_left]) * x
    # accumulate integrand values (add y[j-1] to y[j])
    np.cumsum(y, axis=axis, out=y)
    
    return y

def add_noise(arr: np.ndarray, rng, eps: float) -> np.ndarray:
    """
    Creates a noised version of a given data array.
    See [Grathwohl, Kunstmann, Rieder (2018) - Approximate inverse for the common offsetacquisition geometry in 2D seismic imaging] (p.15)

    Parameters
    ----------
        arr : numpy-array
            the base array to add noise to
        rng : numpy.random Generator
            the random number generator
        eps : float
            noise level

    Returns
    ----------
        noised : numpy-array
            noised array
    """
    noise = rng.uniform(-1, 1, size=arr.shape)
    return arr + eps * np.sum(arr**2)**0.5 * noise / np.sum(noise**2)**0.5


from scipy.interpolate import interp1d
from scipy.integrate import cumulative_trapezoid
def weigthed_space(x: np.ndarray, weights: Union[np.ndarray, Callable[[np.ndarray], np.ndarray]], order=1, random=False, plot_weights=False):
    """Draw samples with spacing specified by a density function/array.
    
    Parameters
    ----------
        x : ndarray with shape (n,)
            original grid points, ordered by increasing values.
        weigths : ndarray or function
            corresponding densities (not normalized). A function must accept 'x' normalized to [0, 1].
        order : int (optional)
            interpolation order (1 or 2). Order 2 will require dense sampling and a smoothly varying density to work correctly.
        random : bool (optional)
            whether to return random samples, ignoring endpoint.

    Returns
    ----------
        ndarray shape (n,)
        weighted grid points with values from x[0] to x[-1]
    """
    if callable(weights):
        m, M = np.min(x), np.max(x)
        weights = weights((x - m) / (M - m))
    
    if plot_weights:
        import matplotlib.pyplot as plt
        plt.figure()
        plt.plot(x, weights)
        plt.show()

    cps = cumulative_trapezoid(weights, x, initial=0)
    # normalize
    cps *= (1 + sys.float_info.epsilon) / cps[-1]

    intfunc = interp1d(cps, x, kind=order)
    if random:
        return intfunc(np.random.uniform(size=x.shape))
    else:
        return intfunc(np.linspace(0, 1, len(x)))
