import functools
from multiprocessing import RLock, RawArray, Event, Lock
from multiprocessing.pool import AsyncResult
from typing import Generator, Sequence, Tuple
import numpy as np
import scipy.sparse as sp


def array_to_shared(arr: np.ndarray) -> RawArray:
    """
    Creates a shared-memory RawArray of a 1D-array and copies the data of 'arr' into it.
    """
    assert arr.ndim == 1, "array must be 1D."
    return RawArray(np.ctypeslib.as_ctypes_type(arr.dtype), arr)

def np_to_shared_info(arr: np.ndarray) -> Tuple[RawArray, Sequence[int]]:
    """
    Creates a shared-memory RawArray and copies the data of 'arr' into it.
    """
    arr_shared = RawArray(np.ctypeslib.as_ctypes_type(arr.dtype), arr.size)
    # Wrap X as an numpy array so we can easily manipulates its data.
    arr_np = np.frombuffer(arr_shared).reshape(arr.shape)
    # Copy data to our shared array.
    np.copyto(arr_np, arr)

    return arr_shared, arr.shape

def create_shared(shape: Sequence[int], dtype: np.dtype) -> Tuple[np.ndarray, Tuple[RawArray, Sequence[int]]]:
    """
    Create a new shared array. Return a NumPy array view to the shared array and the info containing the shared array pointer and shape.
    The buffer/array values are not initialized.
    """
    arr_shared = RawArray(np.ctypeslib.as_ctypes_type(dtype), int(np.prod(shape)))
    arr_np = np.frombuffer(arr_shared).reshape(shape)
    return arr_np, (arr_shared, shape)

def np_from_shared(arr_info: Tuple[RawArray, Sequence[int]]) -> np.ndarray:
    """
    Returns a NumPy array view of a shared array.
    """
    return np.frombuffer(arr_info[0]).reshape(arr_info[1])

def sp_to_shared_info(arr: sp.csr_matrix) -> Tuple[RawArray, RawArray, RawArray, Sequence[int]]:
    """
    Creates a sharable info from a sparse matrix.
    """
    assert isinstance(arr, sp.csr_matrix), "Matrix must be in csr format"
    data = RawArray(np.ctypeslib.as_ctypes_type(arr.data.dtype), arr.data)
    indices = RawArray(np.ctypeslib.as_ctypes_type(arr.indices.dtype), arr.indices)
    indptr = RawArray(np.ctypeslib.as_ctypes_type(arr.indptr.dtype), arr.indptr)
    return data, indices, indptr, arr.shape

def sp_from_shared(arr_info: Tuple[RawArray, RawArray, RawArray, Sequence[int]]) -> sp.csr_matrix:
    """
    Returns a sparse array from a shared array info.
    """
    data, indices, indptr, shape = arr_info
    return sp.csr_matrix((data, indices, indptr), shape=shape, copy=False)


class _DummyFuture:
    def __init__(self, res):
        self._res = res

    def ready(self):
        return True
    
    def get(self):
        return self._res

class DummyPool:
    """Single threaded version of the multiprocessing.Pool. Useful for debugging/profiling."""
    def __init__(self, initializer=None, initargs=None):
        if initializer is not None:
            initializer(*initargs)
        
        self._processes = 1
    
    def apply_async(self, func, args):
        res = func(*args)
        return _DummyFuture(res)
    
    def close(self):
        pass

    def join(self):
        pass

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass

    def map(func, *iterables) -> map:
        map(func, *iterables)

def as_completed(futures: Sequence[AsyncResult]) -> Generator[AsyncResult, None, None]:
    """
    An iterator over the given futures that yields each as it completes (multiprocessing version of 'concurrent.futures.as_completed').

    Args:
        futures: The sequence of Futures (possibly created by different Pools) to
            iterate over.

    Returns:
        An iterator that yields the given Futures as they complete (finished or cancelled).
        If any given Futures are duplicated, they will be returned once.
    """
    event = Event()
    lock = RLock()

    num_fs = len(futures)
    finished = set()

    def _set(self, i, obj):
        with lock:
            finished.add(self)
            AsyncResult._set(self, i, obj)
            event.set()

    for future in futures:
        # override the _set method of the result object
        future._set = functools.partial(_set, future)

        # make sure the status doesn't change while checking 'future.ready()'
        with lock:
            if future.ready():
                finished.add(future)
                event.set()

    collected = 0
    while collected < num_fs:
        event.wait(timeout=10) # use timeout of 10s as failsafe

        with lock:
            to_yield = finished
            event.clear()
            finished = set()

        yield from to_yield
        collected += len(to_yield)
