import functools
from typing import Any, Callable, Optional, Sequence
import matplotlib.pyplot as plt
import math
import numpy as np

from matplotlib.lines import Line2D
from matplotlib.collections import QuadMesh
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable


_mods_pressed = {
    'shift': False,
    'control': False,
    'alt': False
}
def _process_key(pressed, event):
    if event.key is None:
        return

    if "shift" in event.key:
        _mods_pressed["shift"] = pressed
    if "control" in event.key or "ctrl" in event.key:
        _mods_pressed["control"] = pressed
    if "alt" in event.key:
        _mods_pressed["alt"] = pressed

def _process_scroll(event):
    fig = event.canvas.figure

    step = int(math.copysign(1, -event.step))
    if _mods_pressed['control'] and fig.n_frames > 100:
        step *= 100
    elif _mods_pressed['shift'] and fig.n_frames > 10:
        step *= 10

    fig.index = (fig.index + step) % fig.n_frames
    fig.suptitle(f"{fig.index+1}/{fig.n_frames}")

    for ax in fig.axes:
        if not hasattr(ax, "frames"):
            break

        ax.set_data(ax.frames[fig.index])

    fig.canvas.draw()

# supported plot types
def _plot_set_data(ax, line, autoscale):
    def set_data(y):
        line.set_ydata(y)

        if autoscale:
            # recompute the ax.dataLim
            ax.relim()
            # update ax.viewLim using the new dataLim
            ax.autoscale_view()
    return set_data

def _pcolormesh_set_data(ax, quad, autoscale):
    def set_data(arr):
        quad.set_array(arr)
        if autoscale:
            quad.autoscale()
    return set_data

_set_data = {
    Line2D: _plot_set_data,  # standard plot
    QuadMesh: _pcolormesh_set_data # pcolormesh
}

def multi_slice_viewer(x: Sequence[Any], y: Sequence[Sequence[Any]], cols=2, autoscale=True, add_subplots=0,
                        plot_funcs=(lambda ax, x, y: ax.plot(x, y)[0],)) -> None:
    """
    Example:
        x1 = np.linspace(0, 1, 101)
        x2 = np.linspace(-1, 1, 101)

        y1 = [x1**i for i in range(10)]
        y2 = [x2 - i for i in range(10))]

        multi_slice_viewer((x1, x2), (y1, y2))
    
    If x-axis is the same for all y-data, we can also write:
        multi_slice_viewer((x1,), (y1, y2))
    """
    # only single x-array provided -> multiply it to match the number of y-arrays
    if len(x) == 1:
        x = list(x) * len(y)
    if len(plot_funcs) == 1:
        plot_funcs = list(plot_funcs) * len(y)
    
    assert len(x) == len(y), f"x and y data must be of same length (x:{len(x)} - y:{len(y)})"
    n_subplots = len(x) + add_subplots

    cols = min(cols, n_subplots)
    rows = math.ceil(n_subplots / cols)

    fig, axes = plt.subplots(nrows=rows, ncols=cols, squeeze=False)
    axes = axes.flatten()[:n_subplots] # it's possible that rows*cols != n_subplots

    fig.index = 0
    fig.autoscale = autoscale
    fig.n_frames = len(y[0]) # assumed, that all data arrays have the same number of frames
    fig.suptitle(f"{fig.index+1}/{fig.n_frames}")

    for ax, x_data, y_frames, plot_func in zip(axes, x, y, plot_funcs):
        assert len(y_frames) == fig.n_frames, f"number of frames must be equal for all y-data."

        ax.frames = y_frames
        obj = plot_func(ax, x_data, y_frames[fig.index])
        ax.set_data = _set_data[type(obj)](ax, obj, autoscale)

    fig.canvas.mpl_connect("scroll_event", _process_scroll)
    fig.canvas.mpl_connect("key_press_event", functools.partial(_process_key, True))
    fig.canvas.mpl_connect("key_release_event", functools.partial(_process_key, False))
    return axes

def add_colorbar(ax, plot, size="3%", pad=0.2):
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size=size, pad=pad)
    return plt.colorbar(plot, cax=cax)


if __name__ == "__main__":
    x1 = np.arange(10)
    y1 = np.array([[i**j for i in range(10)] for j in range(3)])

    x2 = np.arange(5)
    y2 = np.array([[i+j for i in range(5)] for j in range(3)])
    multi_slice_viewer((x1, x2), (y1, y2))
    plt.show()
    # multi_slice_viewer((x1,), (y1,))
    # plt.show()

    X = np.mgrid[0:5:0.1, 0:5:0.1]
    y = [X[0] - i for i in range(5)]

    def plot_func(ax, x, y):
        res = ax.pcolormesh(*x, y, shading="auto")
        plt.colorbar(res, ax=ax)
        return res

    multi_slice_viewer((X,), (y,), plot_func=plot_func)
    plt.show()
