import cProfile
import pstats
import random
import time
from typing import Callable, Tuple
from util import reshelper

_prof = None

class Timer:
    def __init__(self, perf=False):
        self._time_func = time.perf_counter if perf else time.time
        self._start = self._time_func()

    def reset(self):
        self._start = self._time_func()

    def __str__(self) -> str:
        return f"Time elapsed: {self._time_func() - self._start:.2f} seconds"

def start():
    global _prof

    if _prof is not None:
        raise Exception("Profiling already running!")

    _prof = cProfile.Profile()
    _prof.enable()

def stop(filename):
    global _prof

    if _prof is None:
        raise Exception("Profiling needs to be started first!")
    _prof.disable()
    _prof.dump_stats(reshelper.res_path("profiling") + filename + ".prof")
    _prof = None

def print_stats(filename, sort="time", limit=15):
    stat = pstats.Stats(reshelper.res_path("profiling") + filename + ".prof")
    stat.strip_dirs()
    stat.sort_stats(sort)
    stat.print_stats(limit)

"""
prof = cProfile.Profile()
prof.enable()

Code to profile ...

prof.disable()
prof.dump_stats("NAME.prof")
"""

import numpy as np
import scipy.sparse as sp
import sys
import itertools


def get_sizeof(obj, handlers={}, overwrites={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    There is also 'overwrites'. Add classes to this to overwrite the calculation for their object-size (default is sys.getsizeof)

        overwrites = { SomeClass: lambda a: a.size }
    """
    dict_handler = lambda d: itertools.chain.from_iterable(d.items())
    all_handlers = { tuple: iter,
                     list: iter,
                     dict: dict_handler,
                     set: iter,
                     frozenset: iter,
                   }
    all_handlers.update(handlers)     # user handlers take precedence

    all_overwrites = {
        np.ndarray: lambda arr: arr.nbytes, # missing memory from non-array elements but meh
        sp.spmatrix: lambda arr: arr.data.nbytes + arr.indices.nbytes + arr.indptr.nbytes
    }
    all_overwrites.update(overwrites)

    seen = set()                      # track which object id's have already been seen
    default_size = sys.getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))

        for typ, overwrite in all_overwrites.items():
            if isinstance(o, typ):
                s = overwrite(o)
                if verbose:
                    print(s, type(o), repr(o), file=sys.stderr)
                return s
        s = sys.getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=sys.stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(obj)

def sizeof_str(obj, precision=3, verbose=False):
    s = get_sizeof(obj, verbose=verbose)

    n = 0
    while s >= 1024 and n < 5:
        s /= 1024
        n += 1
    
    return f"{s:.{precision}f} " + ("B", "KB", "MB", "GB", "TB")[n]

def run_test(funcs: Tuple[Callable], args: Tuple, n: int=100) -> None:
    """Runs all functions in 'funcs' n times (in random order) and prints out a summary of the results.
    Each function must accept the same arguments 'args'."""
    k = len(funcs)
    order = list(range(k*n))
    random.shuffle(order)

    times = [[] for _ in range(k)]
    for o in order:
        for i in range(k):
            if o < (i+1) * n:
                f_i = i
                break
        start = time.perf_counter()
        funcs[f_i](*args)
        times[f_i].append(time.perf_counter() - start)

    print(f"Test results for {n} iterations.")
    for i, f in enumerate(funcs):
        name = getattr(f, "__name__", str(f))
        print(f"""'{name}':
    total: {np.sum(times[i])}
    avg:   {np.mean(times[i])}""")
    return times


if __name__ == "__main__":
    arrs = [np.eye(10000, 1000, dtype=np.double) for _ in range(10)]
    arrs_sp = [sp.csr_matrix(arr) for arr in arrs]
    print(sizeof_str(arrs, verbose=False))
    print(sizeof_str(arrs_sp, verbose=False))
